# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

import os.path

import pkg_resources
from alembic import command
from alembic import config
from nagare.admin import util
from nagare.admin.command import Command


class AlembicRevisionCommand(Command):
    desc = 'Create a new alembic revision file.'

    @staticmethod
    def set_options(optparser):
        optparser.add_option(
            '-m',
            '--message',
            dest='message',
            default=None,
            help="Message string to use with 'revision'"
        )

    @staticmethod
    def run(parser, options, args):
        cfg = config.Config()
        cfg.set_main_option(
            'script_location',
            os.path.join(
                pkg_resources.require('kansha')[0].location,
                'kansha/alembic'
            )
        )
        command.revision(cfg, message=options.message)


class AlembicUpgradeCommand(Command):
    desc = 'Upgrade to a later alembic version.'

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' revision [application]'

    @staticmethod
    def run(parser, options, args):
        if len(args) not in (1, 2):
            parser.error('Bad number of parameters')

        # Read the configuration of the application
        try:
            application = args[1]
        except IndexError:
            application = 'kansha'

        cfg = _build_alembic_config()
        _set_sqlalchemy_uri(cfg, application, parser.error)

        command.upgrade(cfg, args[0])


class AlembicDowngradeCommand(Command):
    desc = 'Revert to a previous version.'

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' revision [application]'

    @staticmethod
    def run(parser, options, args):
        if len(args) not in (1, 2):
            parser.error('Bad number of parameters')

        # Read the configuration of the application
        try:
            application = args[1]
        except IndexError:
            application = 'kansha'

        cfg = _build_alembic_config()
        _set_sqlalchemy_uri(cfg, application, parser.error)

        command.downgrade(cfg, args[0])


class AlembicCurrentCommand(Command):
    desc = "Display alembic's current revision."

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' [application]'
        optparser.add_option(
            '-v',
            '--verbose',
            action='store_const',
            const=True,
            default=False,
            dest='verbose',
            help='Use more verbose output'
        )

    @staticmethod
    def run(parser, options, args):
        if len(args) not in (0, 1):
            parser.error('Bad number of parameters')

        # Read the configuration of the application
        try:
            application = args[0]
        except IndexError:
            application = 'kansha'

        cfg = _build_alembic_config()
        _set_sqlalchemy_uri(cfg, application, parser.error)

        command.current(cfg, verbose=options.verbose)


class AlembicStampCommand(Command):
    desc = "'stamp' alembic's revision table with the given revision; don't run any migrations."

    @staticmethod
    def set_options(optparser):
        optparser.usage += ' revision [application]'
        optparser.add_option(
            '-v',
            '--verbose',
            action='store_const',
            const=True,
            default=False,
            dest='verbose',
            help='Use more verbose output'
        )

    @staticmethod
    def run(parser, options, args):
        if len(args) not in (1, 2):
            parser.error('Bad number of parameters')

        # Read the configuration of the application
        try:
            application = args[1]
        except IndexError:
            application = 'kansha'

        cfg = _build_alembic_config()
        _set_sqlalchemy_uri(cfg, application, parser.error)

        command.stamp(cfg, args[0])


def _set_sqlalchemy_uri(cfg, application, error):
    aconf = util.read_application(application, error)[-1]
    cfg.set_main_option('sqlalchemy.url', aconf['database']['uri'])


def _build_alembic_config():
    cfg = config.Config()
    cfg.set_main_option(
        'script_location',
        os.path.join(
            pkg_resources.require('kansha')[0].location,
            'kansha/alembic'
        )
    )
    return cfg



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import presentation, var, security, ajax

from .comp import CardDescription


@presentation.render_for(CardDescription)
def render(self, h, comp, *args):
    """Render description component.

    If text is empty, display a text area otherwise it's a simple text
    """
    id_ = h.generate_id()
    kw = {'class': 'description-content', 'id': id_}
    if security.has_permissions('edit', self.card):
        kw['onclick'] = h.a.action(comp.becomes, model='edit').get('onclick')
    with h.div(**kw):
        if self.text:
            h << h.parse_htmlstring(self.text, fragment=True)
        elif security.has_permissions('edit', self.card):
            h << h.textarea(placeholder=_("Add a description. URL in the text are automatically recognized as links."))

    h << h.script("YAHOO.kansha.app.urlify($('#' + %s))" % ajax.py2js(id_))
    h << h.script('if (typeof editor != "undefined") { editor.destroy(); editor = undefined; }')
    return h.root


def change_text(description, comp, text):
    description.change_text(text)
    comp.becomes(model=None)


@presentation.render_for(CardDescription, model='edit')
def render_edit(self, h, comp, *args):
    """Render description component in edit mode"""
    text = var.Var(self.text)
    with h.div(id_="description", style='visibility: hidden'):
        with h.form(class_='description-form'):
            txt_id = h.generate_id()
            with h.div(id_=txt_id + '-toolbar'):
                h << h.a(h.i(class_='icon-bold'), data_wysihtml_command='bold',
                         class_='btn icon-btn', title=_('bold')) << ' '
                h << h.a(h.i(class_='icon-italic'), data_wysihtml_command='italic',
                         class_='btn icon-btn', title=_('italic')) << ' '
                h << h.a(h.i(class_='icon-underline'), data_wysihtml_command='underline',
                         class_='btn icon-btn', title=_('underline')) << ' '
                h << h.a(h.i(class_='icon-list-numbered'), data_wysihtml_command='insertOrderedList',
                         class_='btn icon-btn', title=_('numbered list')) << ' '
                h << h.a(h.i(class_='icon-list2'), data_wysihtml_command='insertUnorderedList',
                         class_='btn icon-btn', title=_('bullet list')) << ' '
                h << h.a(h.i(class_='icon-link'), data_wysihtml_command='createLink',
                         class_='btn icon-btn', title=_('link')) << ' '
                with h.div(data_wysihtml_dialog='createLink', class_='editor-popin',
                           style='display:none'):
                    with h.label:
                        h << h.input(data_wysihtml_dialog_field='href', value='http://',
                                     type='text', class_='text')
                    h << h.a(_('Link'), data_wysihtml_dialog_action='save', class_='btn btn-primary')
                    h << h.a(_('Cancel'), data_wysihtml_dialog_action='cancel', class_='btn')
            h << h.textarea(
                text(), id_=txt_id,
                class_='description-editor description-content').action(text)

            with h.div(class_='buttons'):
                h << h.button(_('Save'), class_='btn btn-primary').action(lambda: change_text(self, comp, text()))
                h << ' '
                h << h.button(_('Cancel'), class_='btn').action(change_text, self, comp, None)
    h << h.script("""
        var editor = new wysihtml.Editor(%s, { // id of textarea element
          toolbar:      %s, // id of toolbar element
          parserRules:  wysihtmlParserRules, // defined in parser rules set
          style: true,  // adopt styles of main page
          // FIXME: no hard coded path to theme; extensions should have a mechanism to load their styles
          stylesheets: ['/static/kansha/css/themes/kansha_flat/board.css'] // adoption is not complete
        });
        document.getElementById("description").style.visibility = "visible";
        editor.focus();
        """ % (ajax.py2js(txt_id), ajax.py2js(txt_id + '-toolbar'))
    )
    return h.root


@presentation.render_for(CardDescription, model='badge')
def render_badge(self, h, *args):
    """Render description component as a card badge"""
    if self.text:
        with h.span(class_='badge'):
            h << h.span(h.i(class_='icon-pencil'), class_='label')
    return h.root



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.services import service, services


class Service(service.Service):
    CATEGORY = 'service'


class ServicesRepository(services.Services):
    ENTRY_POINTS = 'kansha.services'
    CONFIG_SECTION = 'services'



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--
from collections import OrderedDict

from nagare import i18n
from nagare import editor
from nagare.i18n import _
from nagare import security
from nagare import component
from nagare import validator, var

from kansha import notifications
from kansha.menu import MenuEntry
from kansha import title


class BoardConfig(object):

    """Board configuration component"""

    def __init__(self, board):
        """Initialization

        In:
            - ``board`` -- the board object will want to configure
        """
        self.board = board
        self.menu = OrderedDict()
        self.menu['profile'] = MenuEntry(_(u'Profile'), 'icon-profile', BoardProfile)
        if security.has_permissions('manage', self.board):
            self.menu['labels'] = MenuEntry(_(u'Card labels'), 'icon-price-tag', BoardLabels)
            self.menu['weights'] = MenuEntry(_(u'Card weights'), 'icon-meter', BoardWeights)
            self.menu['background'] = MenuEntry(_(u'Background'), 'icon-paint-format', BoardBackground)
        self.selected = None
        self.content = component.Component(None)
        self.select(self.menu.iterkeys().next())

    def select(self, v):
        """Select a configuration menu item

        In:
            - ``v`` -- the id_ of the menu item we want to show
        """
        self.selected = v
        self.content.becomes(self.menu[v].content(self.board))


class BoardLabels(object):

    """Board configuration component for board labels"""

    def __init__(self, board):
        """Initialization

        In:
            - ``board`` -- the board object we are working on
        """
        self.board = board
        self.labels = []
        for label in board.labels:
            t = component.Component(title.EditableTitle(label.get_title))
            t.on_answer(label.set_title)
            l = component.Component(label, model='edit-color')
            self.labels.append((t, l))


class BoardWeights(object):

    """Board configuration component for card weights"""

    def __init__(self, board):
        """Initialization

        In:
            - ``board`` -- the board object we are working on
        """
        self.board = board
        self._weights_editor = component.Component(WeightsSequenceEditor(self))

    @property
    def weights(self):
        return self.board.weights

    @weights.setter
    def weights(self, value):
        self.board.weights = value

    def activate_weighting(self, weighting_type):
        self.board.activate_weighting(weighting_type)


class WeightsSequenceEditor(editor.Editor):

    """Weights Sequence form"""
    fields = {'weights'}

    def __init__(self, target, *args):
        """Initialization

        In:
            - ``board`` -- the board object we are working on
        """
        super(WeightsSequenceEditor, self).__init__(target, self.fields)
        self.weights.validate(self.validate_sequence)
        self.feedback = ''

    def validate_sequence(self, value):
        try:
            res = validator.StringValidator(value, strip=True).not_empty(msg=i18n._("Required field")).to_string()
        except ValueError:
            raise
        if res:
            try:
                weights = res.split(',')
                for weight in weights:
                    res = validator.IntValidator(weight).greater_or_equal_than(0).to_int()
            except ValueError:
                raise ValueError(i18n._('Must be positive integers'))
        return value

    def commit(self):
        self.feedback = ''
        if self.is_validated(self.fields):
            super(WeightsSequenceEditor, self).commit(self.fields)
            self.feedback = i18n._('Sequence saved')
            return True
        return False


class BoardProfile(object):

    """Board configuration component"""

    def __init__(self, board):
        """Initialization

        In:
            - ``board`` -- the board object we are working on
        """
        self.board = board

    def allow_comments(self, v):
        """Changes comments permissions

        3 values allowed (off/member/public)

        In:
            - ``v`` -- an integer (see security.py for values)
        """
        self.board.allow_comments(v)

    def allow_votes(self, v):
        """Changes votes permissions

        3 values allowed (off/member/public)

        In:
            - ``v`` -- an integer (see security.py for values)
        """
        self.board.allow_votes(v)

    def allow_notifications(self, v):
        notifications.allow_notifications(security.get_user(), self.board, v)

    @property
    def notifications_allowed(self):
        return notifications.notifications_allowed(security.get_user(), self.board)

    def set_archive(self, value):
        self.board.show_archive = value


class BoardBackground(object):

    """Configuration component for background customization
    """

    def __init__(self, board):
        """Initialization

        In:
            - ``board`` -- the board object we are working on
        """
        self.board = board
        self.background_position = var.Var(self.board.background_image_position)

    def set_background_position(self):
        self.board.set_background_position(self.background_position())

    def set_background(self, img):
        self.board.set_background_image(img)

    def reset_background(self):
        self.board.set_background_image(None)

    def set_color(self, color):
        self.board.set_title_color(color)

    def reset_color(self):
        self.board.set_title_color('')



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare import ajax
from nagare import presentation, component

from kansha.events import PopinClosed, EventHandlerMixIn


#TODO: replace Popin by Modal everywhere.

class Popin(EventHandlerMixIn):

    """Popin component"""

    panel_id = 0

    def __init__(self, comp, model):
        """Initialization

        In:
            - ``comp`` -- the component wrapped by the popin
            - ``model`` -- the model of the wrapped component used
                to populate the popin content
        """
        self.comp = component.Component(comp)
        self.model = model

        Popin.panel_id += 1
        self.id = 'panel%d' % Popin.panel_id

    def get_business_object(self):
        business_obj = self.comp()
        while isinstance(business_obj, Popin):
            # nested popins can happen when the user clicks too fast
            self.comp = business_obj.comp
            business_obj = self.comp()
        return business_obj

@presentation.render_for(Popin, 'dnd')
@presentation.render_for(Popin, 'edit')
def render(self, h, comp, *args):
    return h.root


@presentation.render_for(Popin, 'calendar')
@presentation.render_for(Popin)
def render(self, h, comp, model):
    """Render the popin and opens it"""
    business_obj = self.get_business_object()
    emit_event = getattr(business_obj, 'emit_event', self.emit_event)
    action = h.a.action(emit_event, comp, PopinClosed, comp).get('onclick')
    close_func = 'function (){%s;}' % (action)

    h << self.comp.on_answer(lambda x: None).render(h, model=model) # Ignore answers from this view

    with h.div(style='display: none'):
        with h.div(id=self.id):
            self.comp.on_answer(self.handle_event, comp)  # Popin is transparent
            h << self.comp.render(h.AsyncRenderer(), model=self.model)

    h << h.script(
        "YAHOO.kansha.app.showPopin(%s, %s)" % (ajax.py2js(self.id), close_func)
    )

    return h.root


class Modal(object):
    def __init__(self, comp, force_refresh=False):
        if not isinstance(comp, component.Component):
            comp = component.Component(comp)
        self.comp = comp
        self.force_refresh = force_refresh
        self.plugged = False

    def plug(self, comp):
        if not self.plugged:
            self.comp.on_answer(comp.answer)
            self.plugged = True


@presentation.render_for(Modal)
def render_PopinV2(self, h, comp, model):
    self.plug(comp)
    with h.div(class_='modal', id_='modal'):
        with h.div:
            if self.force_refresh:
                h << self.comp.render(h.SyncRenderer())
            else:
                h << self.comp
    if self.force_refresh:
        h << h.script('''$("#modal").click(function(e) {
        if ($(e.target).hasClass('modal')) {
            window.location = "%s";
        }
    });''' % h.SyncRenderer().a.action(comp.answer).get('href'))
    else:
        h << h.script('''$("#modal").click(function(e) {
        if ($(e.target).hasClass('modal')) {
           %s;
        }
    });''' % h.a.action(comp.answer).get('onclick'))
    return h.root


class Empty(object):

    """A component showing only a bit of Javascript
    """

    def __init__(self, js=None):
        """Initialization

        In:
            - ``js`` -- the javascript code to render
        """
        self.js = js


@presentation.render_for(Empty)
def render(self, h, comp, *args):
    """Render the js code"""
    if self.js is not None:
        h.head.javascript(h.generate_id(), self.js)
    h << ' '
    return h.root



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from datetime import datetime

from sqlalchemy import func
from elixir import using_options
from elixir import ManyToOne, OneToMany
from elixir import Field, Unicode, Integer, Boolean
from sqlalchemy.ext.orderinglist import ordering_list

from nagare.database import session

from kansha.card.models import DataCard
from kansha.models import Entity


class DataColumn(Entity):
    """Column mapper
    """
    using_options(tablename='column')
    title = Field(Unicode(200))
    index = Field(Integer)
    nb_max_cards = Field(Integer)
    archive = Field(Boolean, default=False)
    cards = OneToMany('DataCard', order_by='index',  # cascade='delete',
                      collection_class=ordering_list('index'), lazy='subquery')
    board = ManyToOne('DataBoard', colname='board_id')

    def update(self, other):
        self.title = other.title
        self.index = other.index
        self.nb_max_cards = other.nb_max_cards
        session.flush()

    @classmethod
    def create_column(cls, board, index, title, nb_cards=None, archive=False):
        """Create new column

        In:
            - ``board`` -- DataBoard, father of the column
            - ``index`` -- position in the board
            - ``title`` -- title of the column
        Return:
            - created DataColumn instance
        """
        q = cls.query
        q = q.filter(cls.index >= index)
        q = q.filter(cls.board == board)
        q.update({'index': cls.index + 1})
        col = cls(title=title, index=index, board=board, nb_max_cards=nb_cards, archive=archive)
        # session.add(col)
        session.flush()
        return col

    def create_card(self, title, user):
        card = DataCard(title=title, creation_date=datetime.now())
        self.cards.append(card)
        session.flush()
        return card

    def remove_card(self, card):
        if card in self.cards:
            self.cards.remove(card)

    def insert_card(self, index, card):
        done = False
        if card not in self.cards:
            self.cards.insert(index, card)
            session.flush()
            done = True
        return done

    def delete_card(self, card):
        self.remove_card(card)
        card.delete()

    def purge_cards(self):
        for card in self.cards:
            card.delete()

    def append_card(self, card):
        done = False
        if card not in self.cards:
            card.index = None
            self.cards.append(card)
            session.flush()
            done = True
        return done

    @classmethod
    def delete_column(cls, column):
        """Delete column

        Delete a given column, re-index other columns and delete all cards
        of the column

        In:
            - ``column`` -- DataColumn instance to delete
        """
        index = column.index
        board = column.board
        column.delete()
        session.flush()
        q = cls.query
        q = q.filter(cls.index >= index)
        q = q.filter(cls.board == board)
        q.update({'index': cls.index - 1})

    def get_cards_count(self):
        q = DataCard.query.filter(DataCard.column_id == self.id)
        return q.with_entities(func.count()).scalar()



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _

from kansha import validator
from kansha.board import excel_export
from kansha.validator import clean_text
from kansha.services.search import schema
from kansha.cardextension import CardExtension

from .models import DataCardDescription


class CardDescription(CardExtension):
    """Description component,

    This component can be used for with all element which have a description
    field.
    Examples are available in card and board modules
    """

    LOAD_PRIORITY = 20

    def __init__(self, card, action_log, configurator):
        """Initialization

        In:
            - ``card`` -- the card
        """
        super(CardDescription, self).__init__(card, action_log, configurator)

    @staticmethod
    def get_schema_def():
        return schema.Text(u'description')

    def update_document(self, document):
        desc = self.text
        document.description = clean_text(desc) if desc else u''

    def update(self, other):
        self.data.update(other.data)

    @property
    def data(self):
        data = DataCardDescription.get_by_card(self.card.data)
        if data is None:
            data = DataCardDescription.new(self.card.data)
        return data

    @property
    def text(self):
        return self.data.description

    @text.setter
    def text(self, text):
        self.data.description = text

    def change_text(self, text):
        """Edit the description

        In:
            - ``text`` -- the text of the description
        """
        # historically, canceling the description edition calls this function with None.
        # FIXME: that should not happen.
        if text is None:
            return
        text = text.strip()
        if text:
            text = validator.clean_html(text)
        self.text = text

    def __nonzero__(self):
        """Return False if the description if empty
        """
        return bool(self.text)

    def delete(self):
        self.data.delete()


@excel_export.get_extension_title_for(CardDescription)
def get_extension_title_CardDescription(card_extension):
    return _(u'Description')


@excel_export.write_extension_data_for(CardDescription)
def write_extension_data_CardDescription(self, sheet, row, col, style):
    text = validator.clean_text(self.text) if self.text else ''
    sheet.write(row, col, text, style)



from __future__ import absolute_import, unicode_literals

from nagare.services import plugin, plugins
from .services_repository import Service


class Authentication(plugin.Plugin):
    CATEGORY = 'authentication'


class AuthenticationsRepository(Service, plugins.Plugins):
    ENTRY_POINTS = 'kansha.authentication'
    CONFIG_SECTION = 'authentication'

    def __init__(self, conf_filename=None, error=None, services_service=None):
        plugins.Plugins.__init__(self, conf_filename, error, self.config)

    def register(self, name, plugin):
        if plugin.config['activated']:
            self[name] = plugin

    # try:
    #     component_conf.update({'root': conf['root']})
    # except KeyError:
    #     pass
    # try:
    #     component_conf.update({'here': conf['here']})
    # except KeyError:
    #     pass



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import ajax, presentation

from .comp import CardWeightEditor


def answer(editor, comp):
    if editor.commit():
        comp.answer()


@presentation.render_for(CardWeightEditor, 'action')
def render_cardweighteditor(self, h, comp, *args):
    if self.weighting_switch:
        h << self.action_button
    return h.root


@presentation.render_for(CardWeightEditor, 'action_button')
def render_CardWeightEditor_action_button(self, h, comp, *args):
    self.weight(str(self.data.weight))
    h << h.a(h.i(class_='icon-star-empty'), self.data.weight, class_='btn').action(
             comp.call, self, model='edit')
    return h.root


@presentation.render_for(CardWeightEditor, 'edit')
def render_CardWeightEditor_edit(self, h, comp, *args):
    if self.weighting_switch == self.WEIGHTING_FREE:
        id_ = h.generate_id('weight')
        with h.form:
            h << h.input(
                value=self.weight(),
                type='text',
                id_=id_).action(self.weight).error(self.weight.error)
            h << h.button(_('Save'), class_='btn btn-primary').action(answer, self, comp)
            h << h.script("""document.getElementById(%s).focus(); """ % ajax.py2js(id_))
    elif self.weighting_switch == self.WEIGHTING_LIST:
        with h.form:
            with h.div(class_='btn select'):
                with h.select.action(self.weight):
                    for value in self.allowed_weights.split(','):
                        h << h.option(value, value=value.strip()).selected(self.weight())
            h << h.button(_('Save'), class_='btn btn-primary').action(answer, self, comp)

    return h.root


@presentation.render_for(CardWeightEditor, 'badge')
def render_CardWeightEditor_badge(self, h, comp, *args):
    if self.data.weight:
        with h.span(class_='badge'):
            h << h.span(h.i(class_='icon-star-empty'), ' ',
                        self.data.weight, class_='label', title=_(u'Weight'))
    return h.root



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--
"""
Abstract Query expressions.
Queries are not tied to a search engine. They can be built independently, persisted and
then executed by any search engine.
Each search engine is responsible for interpreting queries at search time
by calling them with a mapper.

Search engines don't support joins, so you can query only one document type at once.

See unit tests for example usage.
"""


class Query(object):

    operation = 'nop'

    def __init__(self, field, value):
        self.field = field
        self.value = value
        self.target_schema = self.field.schema

    def __call__(self, mapper):
        op = getattr(mapper, self.operation)
        return op(self.field, self.value)

    def search(self, index_cursor, mapper, limit):
        mapped_query = self(mapper)
        fields_to_load = [name for name, field in self.target_schema.iter_fields() if field.stored]
        index_cursor.search(self.target_schema.type_name, fields_to_load, mapped_query, limit)
        return index_cursor.get_results(self.target_schema.delta)

    def __and__(self, other):
        return ANDQuery(self, other)

    def __or__(self, other):
        return ORQuery(self, other)


class EQQuery(Query):
    operation = 'eq'


class NEQuery(Query):
    operation = 'neq'


class GTQuery(Query):
    operation = 'gt'


class GTEQuery(Query):
    operation = 'gte'


class LTQuery(Query):
    operation = 'lt'


class LTEQuery(Query):
    operation = 'lte'


class MATCHQuery(Query):
    operation = 'match'


class MATCHANYQuery(Query):
    '''
    Match any field in the document.
    The first argument is not a field but a document
    '''
    operation = 'matchany'

    def __init__(self, doc, value):
        self.field = doc
        self.value = value
        self.target_schema = doc


class PHRASEQuery(Query):
    operation = 'phrase'


class INQuery(Query):
    '''Passed value must be iterable'''
    operation = 'in_'


class ANDQuery(Query):

    operation = 'and_'

    def __init__(self, q1, q2):
        assert(q1.target_schema == q2.target_schema)
        self.q1 = q1
        self.q2 = q2
        self.target_schema = q1.target_schema

    def __call__(self, mapper):
        op = getattr(mapper, self.operation)
        return op(self.q1(mapper), self.q2(mapper))

    def __and__(self, other):
        return ANDQuery(self, other)

    def __or__(self, other):
        return ORQuery(self, other)


class ORQuery(ANDQuery):

    operation = 'or_'



"""Add column archive flag

Revision ID: 6cb3c668d65
Revises: 36d972f11e3e
Create Date: 2014-03-03 16:07:47.321713

"""

# revision identifiers, used by Alembic.
revision = '6cb3c668d65'
down_revision = '36d972f11e3e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('column', sa.Column('archive', sa.Boolean, server_default='false'))
    op.add_column('board', sa.Column('archive', sa.Integer, server_default='0'))
    op.add_column('board', sa.Column('archived', sa.Boolean, server_default='false'))
    op.add_column('board', sa.Column('weighting_cards', sa.Integer, server_default='0'))
    op.add_column('board', sa.Column('weights', sa.Text, server_default=''))
    op.add_column('card', sa.Column('weight', sa.Text, server_default=''))


def downgrade():
    op.drop_column('column', 'archive')
    op.drop_column('board', 'archive')
    op.drop_column('board', 'archived')
    op.drop_column('board', 'weighting_cards')
    op.drop_column('board', 'weights')
    op.drop_column('card', 'weight')



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

import random

import webob.exc

from nagare import ajax
from nagare import presentation
from nagare.ajax import YUI_PREFIX


class Autocomplete(object):

    """
    Enhance an existing field by providing an auto-completion feature.

    Parameters:
      - ``field_id``: the field (input or textarea) that should be enhanced
      - ``completion_func``: a function that returns the completion suggestions
        for a given query: it could be a collection of name strings or a
        collection of (name, markup) pairs; the markup is shown in the
        drop-down list (or the name when markup is not available) and the name
        is inserted in the field when the item is selected
      - ``delimiter``: specify the delimiter used if the field is multi-valued
    """

    def __init__(self, field_id, completion_func, delimiter=None,
                 min_query_length=3, max_results_displayed=20):
        self.field_id = field_id
        self.completion_func = completion_func
        self.delimiter = delimiter
        self.min_query_length = int(min_query_length)
        self.max_results_displayed = int(max_results_displayed)
        self.var = 'autocomplete' + str(random.randint(10000000, 99999999))

    def _completion_results(self, query, static_url):
        def make_pair(item):
            if hasattr(item, '__iter__'):
                return item
            else:
                return item, item

        results = self.completion_func(query, static_url)
        return [make_pair(item) for item in results]


def json_response(data):
    """Create a JSON response from the data"""
    return webob.exc.HTTPOk(json_body=data)


@presentation.render_for(Autocomplete, model='static_dependencies')
def render_static_dependencies(self, h, comp, *args):
    # Unused for now...
    for mod in ('autocomplete',):
        h.head.css_url(YUI_PREFIX +
                       '/%(mod)s/assets/skins/sam/%(mod)s.css' % dict(mod=mod))

    h.head.javascript_url(YUI_PREFIX + '/yahoo-dom-event/yahoo-dom-event.js')
    for mod in ('connection', 'animation', 'json',
                'datasource', 'autocomplete'):
        h.head.javascript_url(
            YUI_PREFIX + '/%(mod)s/%(mod)s-min.js' % dict(mod=mod))
    h.head.javascript_url('js/autocomplete.js')
    return h.root


@presentation.render_for(Autocomplete)
def render_autocomplete(self, h, comp, *args):

    static_url = h.head.static_url

    def get_results(query):
        raise json_response(self._completion_results(query, static_url))

    h << h.script(
        'var %(var)s = YAHOO.kansha.autocomplete.init(%(field_id)s,'
        ' %(completion_url)s, %(delimiter)s, %(min_query_length)s, '
        '%(max_results_displayed)s)' %
        {
            'var': self.var,
            'field_id': ajax.py2js(self.field_id),
            'completion_url': ajax.py2js(
                h.add_sessionid_in_url(
                    params=(
                        '_a',
                        '%s=' % h.register_callback(1, get_results, False)
                    )
                )
            ),
            'delimiter': ajax.py2js(self.delimiter),
            'min_query_length': self.min_query_length,
            'max_results_displayed': self.max_results_displayed,
        }
    )

    return h.root



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from nagare import var

from .models import DataHistory


class ActionLog(object):

    def __init__(self, board, card=None):
        self._board = board
        self._card = card

       # View API
        self.user_id = var.Var('')
        self.card_id = var.Var(None)

    @property
    def board(self):
        return self._board

    @property
    def card(self):
        return self._card

    def for_card(self, card):
        """Return new action log restricted to the card.
        The board is the same as self."""
        return ActionLog(self._board, card)

    def add_history(self, user, action, data):
        '''user is App User'''
        DataHistory.add_history(
            self.board.data,
            self.card and self.card.data,
            user.data, action, data
        )

    def delete_card(self):
        if not self._card:
            return
        DataHistory.purge(self._card.data)

    def get_events(self, hours=None):
        return DataHistory.get_events(self.board and self.board.data, hours)

    def get_last_activity(self):
        return DataHistory.get_last_activity(self._board.data)

    @staticmethod
    def get_events_for_data(data_board, hours=None):
        return DataHistory.get_events(data_board, hours)

    # view API
    def get_history(self):
        'internal'
        return DataHistory.get_history(self.board.data, cardid=self.card_id(),
                                       username=self.user_id())


class DummyActionLog(ActionLog):

    def __init__(self):
        super(DummyActionLog, self).__init__(None)

    def add_history(self, user, action, data):
        pass

    def get_last_activity(self):
        return []

    def get_history(self):
        return []



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --


# messages

# FIXME: how can extensions use their own .mo files for translations?

from peak.rules import when

from nagare.i18n import _


def render_event(action, data):
    '''Base function to be augmented with peak.rules'''
    return 'Undefined event type "%s"' % action


@when(render_event, "action=='card_create'")
def render_event_card_create(action, data):
    return _(u'User %(author)s has added card "%(card)s" to column "%(column)s"') % data


@when(render_event, "action=='card_delete'")
def render_event_card_delete(action, data):
    return _(u'User %(author)s has deleted card "%(card)s"') % data


@when(render_event, "action=='card_archive'")
def render_event_card_archive(action, data):
    return _(u'User %(author)s has archived card "%(card)s"') % data


@when(render_event, "action=='card_move'")
def render_event_card_move(action, data):
    return _(u'User %(author)s has moved card "%(card)s" from column "%(from)s" to column "%(to)s"') % data


@when(render_event, "action=='card_title'")
def render_event_card_title(action, data):
    return _(u'User %(author)s has renamed card "%(from)s" to "%(to)s"') % data



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import component, editor, i18n, security, var, validator as nagare_validator

from kansha import title
from kansha import events
from kansha import exceptions
from kansha.toolbox import popin, overlay
from kansha.card import Card, NewCard

from .models import DataColumn


class Column(events.EventHandlerMixIn):

    """Column component
    """

    def __init__(self, id_, board, card_extensions, action_log, card_filter,
                 search_engine_service, services_service, data=None):
        """Initialization

        In:
            - ``id_`` -- the id of the column
        """
        self.db_id = id_
        self._data = data
        self.id = 'list_' + str(self.db_id)
        self.board = board
        self._services = services_service
        self.action_log = action_log
        self.card_filter = card_filter
        self.search_engine = search_engine_service
        self.card_extensions = card_extensions
        self.body = component.Component(self, 'body')
        self.title = component.Component(
            title.EditableTitle(self.get_title)).on_answer(self.set_title)
        self.card_counter = component.Component(CardsCounter(self))
        self._cards = None
        self.new_card = component.Component(
            NewCard(self))

    @property
    def cards(self):
        if self._cards is None:
            self._cards = [
                component.Component(
                    self._services(
                        Card, c.id,
                        self.card_extensions,
                        self.action_log,
                        self.card_filter,
                        data=c))
                for c in self.data.cards]
        return self._cards

    def update(self, other):
        self.data.update(other.data)
        cards_to_index = []
        for card_comp in other.cards:
            card = card_comp()
            new_card = self.create_card(card.get_title())
            new_card.update(card)
            cards_to_index.append(new_card)
        self.index_cards(cards_to_index)

    def index_cards(self, cards, update=False):
        for card in cards:
            card.add_to_index(self.search_engine, self.board.id, update=update)
        self.search_engine.commit(True)

    def actions(self, action, comp):
        if action == 'empty':
            self.empty()
        elif action == 'delete':
            self.emit_event(comp, events.ColumnDeleted, comp)
        elif action == 'purge':
            self.purge_cards()
        self.emit_event(comp, events.SearchIndexUpdated)

    def ui_create_card(self, comp, title):
        card = self.create_card(title)
        if card:
            self.index_cards([card])
            self.emit_event(comp, events.SearchIndexUpdated)
            self.card_filter.reset()

    def on_event(self, comp, event):
        if event.is_(events.CardClicked):
            card_comp = event.data
            card_comp.becomes(popin.Popin(card_comp, 'edit'))
        elif event.is_(events.ParentTitleNeeded):
            return self.get_title()
        elif event.is_(events.CardEditorClosed):
            card_bo = event.emitter
            slot = event.data
            slot.becomes(card_bo)
            # if card has been edited, reindex
            if security.has_permissions('edit', card_bo):
                card_bo.add_to_index(self.search_engine, self.board.id, update=True)
                self.search_engine.commit(True)
                self.emit_event(comp, events.SearchIndexUpdated)
            card_bo.refresh()
        elif event.is_(events.CardArchived):
            self.remove_card_by_id(event.last_relay.id)

    @property
    def data(self):
        """Return the column object from the database
        """
        if self._data is None:
            self._data = DataColumn.get(self.db_id)
        return self._data

    def __getstate__(self):
        self._data = None
        return self.__dict__

    def set_title(self, title):
        """Set title

        In:
            - ``title`` -- new title
        """
        self.data.title = title

    def get_title(self):
        """Get title

        Return :
            - the board title
        """
        if self.is_archive:
            return i18n._(u'Archived cards')
        return self.data.title

    def delete(self, purge=False):
        """Delete itself"""
        if not self.data:
            return
        if purge:
            self.purge_cards()
        else:
            self.empty()
        DataColumn.delete_column(self.data)

    def empty(self):
        # FIXME: move this function to board only, use Events
        self.board.archive_cards([card() for card in self.cards], self)
        self._cards = None

    def remove_card_comp(self, card):
        self.cards.remove(card)
        business = card()
        if isinstance(business, popin.Popin):
            business = business.get_business_object()
        self.data.remove_card(business.data)

    def remove_card_by_id(self, card_id):
        """Remove card and return corresponding Component."""
        # find component
        card_comp = filter(lambda x: x().id == card_id, self.cards)[0]
        try:
            self.remove_card_comp(card_comp)
        except (IndexError, ValueError):
            raise ValueError(u'Card has been deleted or does not belong to this list anymore')
        return card_comp

    def insert_card(self, index, card):
        inserted = False
        # TODO: when column extensions are introduced, generalize this
        if self.card_counter().check_add(card) and self.data.insert_card(index, card.data):
            self.cards.insert(index, component.Component(card))
            inserted = True
        return inserted

    def insert_card_comp(self, comp, index, card_comp):
        inserted = False
        card = card_comp()
        # TODO: when column extensions are introduced, generalize this
        if self.card_counter().check_add(card) and self.data.insert_card(index, card_comp().data):
            self.cards.insert(index, card_comp)
            card_comp.on_answer(self.handle_event, comp)
            inserted = True
        return inserted

    def delete_card(self, card):
        """Delete card

        In:
            - ``card`` -- card to delete
        """
        self.cards.pop(card.index)
        values = {'column_id': self.id, 'column': self.get_title(), 'card': card.get_title()}
        card.action_log.add_history(
            security.get_user(),
            u'card_delete', values)
        self.search_engine.delete_document(card.schema, card.id)
        self.search_engine.commit()
        card.delete()
        self.data.delete_card(card.data)

    def purge_cards(self):
        for card_comp in self.cards:
            card = card_comp()
            values = {'column_id': self.id, 'column': self.get_title(), 'card': card.get_title()}
            card.action_log.add_history(
                security.get_user(),
                u'card_delete', values)
            card.delete()
            self.search_engine.delete_document(card.schema, card.id)
        del self.cards[:]
        self.search_engine.commit()
        self.data.purge_cards()

    def append_card(self, card):
        # TODO: when column extensions are introduced, generalize this
        if self.card_counter().check_add(card) and self.data.append_card(card.data):
            self.cards.append(component.Component(card))

    def create_card(self, text=''):
        """Create a new card

        In:
            - ``text`` -- the title of the new card
        """
        if text:
            if not self.can_add_cards:
                raise exceptions.KanshaException(_('Limit of cards reached fo this list'))
            new_card = self.data.create_card(text, security.get_user().data)
            card_obj = self._services(Card, new_card.id, self.card_extensions, self.action_log,
                                      self.card_filter)
            self.cards.append(component.Component(card_obj, 'new'))
            values = {'column_id': self.id,
                      'column': self.get_title(),
                      'card': new_card.title}
            card_obj.action_log.add_history(
                security.get_user(),
                u'card_create', values)
            return card_obj

    def change_index(self, new_index):
        """Change index of the column

        In:
            - ``index`` -- new index
        """
        self.data.index = new_index

    def set_nb_cards(self, nb_cards):
        self.data.nb_max_cards = int(nb_cards) if nb_cards else None

    @property
    def can_add_cards(self):
        rval = True
        if self.nb_max_cards is not None:
            rval = self.count_cards < self.nb_max_cards
        return rval

    @property
    def nb_max_cards(self):
        return self.data.nb_max_cards

    @property
    def count_cards(self):
        return self.data.get_cards_count()

    @property
    def is_archive(self):
        return self.data.archive

    @is_archive.setter
    def is_archive(self, value):
        self.data.archive = value


class NewColumnEditor(object):

    """Column creator component
    """

    def __init__(self, columns_count):
        """Initialization

        In:
            - ``board`` -- the board the new column will belong
        """
        self.columns_count = columns_count
        self.index = editor.Property(u'').validate(nagare_validator.to_int)
        self.title = editor.Property(u'')
        self.title.validate(lambda v: nagare_validator.to_string(v.strip()).not_empty(_(u'''Can't be empty''')))
        self.nb_cards = editor.Property(u'').validate(self.validate_nb_cards)

    def is_validated(self):
        return all((
            self.index.error is None,
            self.title.error is None,
            self.nb_cards.error is None
        ))

    def validate_nb_cards(self, value):
        if value:
            return nagare_validator.to_int(value)
        return value

    def commit(self, comp):
        if self.is_validated():
            comp.answer((self.index.value, self.title.value, self.nb_cards.value))

    def cancel(self, comp):
        comp.answer(None)


# TODO: move data from column to CardsCounter model and make it a column extension
class CardsCounter(object):

    def __init__(self, column):
        self.column = column
        self.id = self.column.id + '_counter'
        self.text = str(self.column.nb_max_cards or 0)
        self.error = None
        self.editable_counter = component.Component(self)

    # Public methods

    def check_add(self, card=None):
        return not self.column.nb_max_cards or self.column.count_cards < self.column.nb_max_cards

    # Private methods

    def change_nb_cards(self, text):
        """Change the title of our wrapped object

        In:
            - ``text`` -- the new title
        Return:
            - the new title

        """

        self.text = text
        self.column.set_nb_cards(text)
        return text

    def reset_error(self):
        self.error = None

    def validate(self, text, comp):
        self.reset_error()
        nb = int(text) if text else 0
        count = self.column.count_cards
        if not nb or nb >= count:
            self.change_nb_cards(nb)
            comp.answer()
        else:
            self.error = _('Must be bigger than %s') % count



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import cgi
import sys
import json
import time
import pstats
import urllib
import urlparse
import cProfile as profile
from collections import OrderedDict

import configobj
import pkg_resources

from nagare.i18n import _, _L
from nagare.admin import command
from nagare.namespaces import xhtml5
from nagare import component, wsgi, security, config, log, i18n

from kansha import events
from kansha.card import Card
from kansha import exceptions
from kansha.menu import MenuEntry
from kansha.authentication import login
from kansha import services, notifications
from kansha.services.search import SearchEngine
from kansha.user.usermanager import UserManager
from kansha.user.user_profile import get_userform  # !!!!!!!!!!!!!!!
from kansha.board.boardsmanager import BoardsManager
from kansha.security import SecurityManager, Unauthorized


def run():
    return command.run('kansha.commands')


class Kansha(object):
    """The Kansha root component"""

    def __init__(self, app_title, app_banner, favicon, theme,
                 card_extensions, services_service):
        """Initialization
        """
        self.app_title = app_title
        self.app_banner = app_banner
        self.favicon = favicon
        self.theme = theme
        self.card_extensions = card_extensions
        self._services = services_service

        self.title = component.Component(self, 'tab')
        self.user_menu = component.Component(None)
        self.content = component.Component(None)
        self.user_manager = UserManager()
        self.boards_manager = self._services(
            BoardsManager, self.app_title, self.app_banner, self.theme,
            card_extensions)

        self.home_menu = OrderedDict()
        self.selected = 'board'

    def _on_menu_entry(self, id_):
        """Select a configuration menu entry

        In:
            - ``id_`` -- the id of the selected menu entry
        """
        if id_ == 'boards':
            self.boards_manager.load_user_boards()
        self.content.becomes(self.home_menu[id_].content)
        self.selected = id_

    def initialization(self):
        """ Initialize Kansha application

        Initialize user_menu with current user,
        Initialize last board

        Return:
         - app initialized
        """
        user = security.get_user()
        self.home_menu['boards'] = MenuEntry(
            _L(u'Boards'),
            'table2',
            self.boards_manager
        )
        self.home_menu['profile'] = MenuEntry(
            _L(u'Profile'),
            'user',
            self._services(
                get_userform(
                    self.app_title, self.app_banner, self.theme, user.data.source
                ),
                user.data,
            )
        )
        self.user_menu = component.Component(user)
        if user and self.content() is None:
            self.select_last_board()
        return self

    def _select_board(self, board):
        self.content.becomes(board)
        # if user is logged, update is last board
        user = security.get_user()
        if user:
            user.set_last_board(board)

    def select_board(self, id_):
        """Redirect to a board by id

        In:
          - ``id_`` -- the id of the board
        """
        if not id_:
            return
        board = self.boards_manager.get_by_id(id_)
        if board is not None and not board.archived:
            self.content.becomes(board, 'redirect')
            self.selected = 'board'
        else:
            raise exceptions.BoardNotFound()

    def select_board_by_uri(self, uri):
        """Selected a board by URI

        In:
          - ``uri`` -- the uri of the board
        """
        if not uri:
            return
        board = self.boards_manager.get_by_uri(uri)
        if board is not None and not board.archived:
            self._select_board(board)
        else:
            raise exceptions.BoardNotFound()

    def select_last_board(self):
        """Selects the last used board if it's possible

        Otherwise, content becomes user's home
        """
        user = security.get_user()
        data_board = user.get_last_board()
        if data_board and not data_board.archived and data_board.has_member(user.data):
            self.select_board(data_board.id)
        else:
            self._on_menu_entry('boards')

    def handle_event(self, event):
        if event.is_(events.BoardLeft) or event.is_(events.BoardArchived):
            return self._on_menu_entry('boards')
        elif event.is_(events.NewTemplateRequested):
            return self.boards_manager.create_template_from_board(event.emitter, *event.data)


class MainTask(component.Task):
    def __init__(self, app_title, theme, config, card_extensions, services_service):
        self.app_title = app_title
        self.theme = theme
        self._services = services_service
        self.app_banner = config['pub_cfg']['banner']
        self.favicon = config['pub_cfg']['favicon']
        self.app = services_service(
            Kansha,
            self.app_title,
            self.app_banner,
            self.favicon,
            self.theme,
            card_extensions,
        )
        self.config = config

    def go(self, comp):
        user = security.get_user()
        while user is None:
            # not logged ? Call login component
            comp.call(
                self._services(
                    login.Login,
                    self.app_title,
                    self.app_banner,
                    self.favicon,
                    self.theme,
                    self.config
                )
            )
            user = security.get_user()
            user.update_last_login()

        comp.call(self.app.initialization())
        # Logout
        if user is not None:
            security.get_manager().logout()


class WSGIApp(wsgi.WSGIApp):
    """This application uses a HTML5 renderer"""
    renderer_factory = xhtml5.Renderer

    ConfigSpec = {
        'application': {'as_root': 'boolean(default=True)',
                        'title': 'string(default="")',
                        'banner': 'string(default="")',
                        'theme': 'string(default="kansha_flat")',
                        'favicon': 'string(default="img/favicon.ico")',
                        'disclaimer': 'string(default="")',
                        'activity_monitor': "string(default='')"},
        'locale': {
            'major': 'string(default="en")',
            'minor': 'string(default="US")'
        }
    }

    def set_config(self, config_filename, conf, error):
        super(WSGIApp, self).set_config(config_filename, conf, error)
        conf = configobj.ConfigObj(
            conf, configspec=configobj.ConfigObj(self.ConfigSpec), interpolation='Template')
        config.validate(config_filename, conf, error)

        self._services = services.ServicesRepository(
            config_filename, error, conf
        )

        self.card_extensions = services.CardExtensions(
            config_filename, error, conf
        )

        self.as_root = conf['application']['as_root']
        self.app_title = unicode(conf['application']['title'], 'utf-8')
        self.app_name = conf['application']['name']
        self.theme = conf['application']['theme']
        self.application_path = conf['application']['path']

        # search_engine engine configuration
        self.search_engine = SearchEngine(**conf['search'])
        self._services.register('search_engine', self.search_engine)
        Card.update_schema(self.card_extensions)

        # Make assets_manager available to kansha-admin commands
        self.assets_manager = self._services['assets_manager']

        # other
        self.security = SecurityManager(conf['application']['crypto_key'])
        self.debug = conf['application']['debug']
        self.default_locale = i18n.Locale(
            conf['locale']['major'], conf['locale']['minor'])
        pub_cfg = {
            'disclaimer': conf['application']['disclaimer'].decode('utf-8'),
            'banner': conf['application']['banner'].decode('utf-8'),
            'favicon': conf['application']['favicon'].decode('utf-8')
        }
        self.app_config = {
            'authentication': conf['authentication'],
            'pub_cfg': pub_cfg
        }
        self.activity_monitor = conf['application']['activity_monitor']

    def set_publisher(self, publisher):
        if self.as_root:
            publisher.register_application(self.application_path, '', self,
                                           self)

    def create_root(self):
        return super(WSGIApp, self).create_root(
            self.app_title,
            self.theme,
            self.app_config,
            self.card_extensions,
            self._services
        )

    def start_request(self, root, request, response):
        super(WSGIApp, self).start_request(root, request, response)
        if security.get_user():
            self.set_locale(security.get_user().get_locale())

    def __call__(self, environ, start_response):
        query = dict(urlparse.parse_qsl(environ['QUERY_STRING'], True))

        if ('state' in query) and (('code' in query) or ('error' in query)):
            state = query.pop('state')

            environ['QUERY_STRING'] = urllib.urlencode(query)
            environ['QUERY_STRING'] += '&' + state
            environ['REQUEST_METHOD'] = 'POST'

            # PATH_INFO cannot be empty or else Nagare will enter in `on_incomplete_url`
            environ['PATH_INFO'] = environ['PATH_INFO'] or '/'
        if self.debug:
            perf = profile.Profile()
            start = time.time()
            ret = perf.runcall(super(WSGIApp, self).__call__, environ, start_response)
            if time.time() - start > 1:
                stats = pstats.Stats(perf)
                stats.sort_stats('cumtime')
                stats.print_stats(60)
            return ret
        else:
            return super(WSGIApp, self).__call__(environ, start_response)

    def on_exception(self, request, response):
        exc_class, e = sys.exc_info()[:2]
        for k, v in request.POST.items():
            if isinstance(v, cgi.FieldStorage):
                request.POST[k] = u'Content not displayed'
        log.exception(e)
        response.headers['Content-Type'] = 'text/html'
        package = pkg_resources.Requirement.parse('kansha')
        error = pkg_resources.resource_string(
            package, 'static/html/error.html')
        error = unicode(error, 'utf8')
        response.charset = 'utf8'
        data = {'text': u'', 'status': 200,
                'go_back': _(u'Go back'),
                'app_title': self.app_title,
                'app_name': self.app_name,
                'theme': self.theme}
        if exc_class == Unauthorized:
            status = 403
            text = _(u"You are not authorized to access this board")
        elif exc_class == exceptions.BoardNotFound:
            status = 404
            text = _(u"This board doesn't exists")
        elif exc_class == exceptions.NotFound:
            status = 404
            text = _(u"Page not found")
        elif exc_class == exceptions.KanshaException:
            status = 500
            text = _(unicode(e.message))
        else:
            raise
        # Raise exception if debug
        if self.debug:
            raise
        data['status'] = status
        data['text'] = text
        response.status = status
        if request.is_xhr:
            response.body = json.dumps({'details': text, 'status': status})
        else:
            response.text = error % data
        return response

    def on_callback_lookuperror(self, request, response, async):
        """A callback was not found

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object
          - ``async`` -- is an XHR request ?
        """
        log.exception("\n%s" % request)
        if self.debug:
            raise

    def send_notifications(self, hours, url):

        mail_sender = self._services['mail_sender']
        # Group users by board
        boards = {}
        for subscriber_bo in notifications.get_subscribers():
            # FIXME: don't use the data directly
            subscriber = subscriber_bo.data
            boards.setdefault(subscriber.board.id, {'board': subscriber.board,
                                                    'subscribers': []})['subscribers'].append(subscriber)

        for board in boards.itervalues():
            if not board['board'].archived:
                events = services.ActionLog.get_events_for_data(board['board'], hours)
                for subscriber in board['subscribers']:
                    data = notifications.filter_events(events, subscriber)
                    if not data:
                        continue
                    locale = UserManager.get_app_user(subscriber.user.username).get_locale()
                    self.set_locale(locale)
                    subject, content, content_html = notifications.generate_email(self.app_title, board['board'],
                                                                                  subscriber.user, hours, url, data)
                    mail_sender.send(subject, [subscriber.user.email], content, content_html)
        if self.activity_monitor:
            events = services.ActionLog.get_events_for_data(None, hours)
            new_users = UserManager.get_all_users(hours)

            if not (events or new_users):
                return
            h = xhtml5.Renderer()
            with h.html:
                h << h.h1('Boards')
                with h.ul:
                    for event in events:
                        notif = event.to_string()
                        if event.card:
                            # IDs are interpreted as anchors since HTML4. So don't use the ID of
                            # the card as a URL fragment, because the browser
                            # jumps to it.
                            ev_url = urlparse.urljoin(url, event.board.url)
                            id_ = '%s#id_card_%s' % (ev_url, event.card.id)
                            notif = h.a(notif, href=id_, style='text-decoration: none;')
                        h << h.li(u'%s : ' % (event.board.title), notif)
                h << h.h1('New users')
                with h.table(border=1):
                    with h.tr:
                        h << h.th('Login')
                        h << h.th('Fullname')
                        h << h.th('Email')
                        h << h.th('Registration date')
                    for usr in new_users:
                        with h.tr:
                            h << h.td(usr.username)
                            h << h.td(usr.fullname)
                            h << h.td(usr.email)
                            h << h.td(usr.registration_date.isoformat())

            mail_sender.send('Activity report for '+url, [self.activity_monitor], u'', h.root.write_htmlstring())


def create_pipe(app, *args, **kw):
    '''Use with Apache only, fixes the content-length when gzipped'''

    try:
        from paste.gzipper import middleware
        app = middleware(app)
    except ImportError:
        pass
    return app


app = WSGIApp(lambda *args: component.Component(MainTask(*args)))



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.services import plugins


class CardExtensions(plugins.Plugins):
    ENTRY_POINTS = 'kansha.card.extensions'
    CONFIG_SECTION = 'card_extensions'
    CONFIGURATORS = {}

    def set_configurators(self, configurators):
        """
        Return a copy of self with
        CONFIGURATORS set.
        """

        repository = self.copy()
        repository.CONFIGURATORS = configurators
        return repository

    def instantiate_items(self, card, action_log, services_service):
        """
        Return items as CardExtension instances for given card.
        """
        return [
            (name, services_service(klass, card, action_log, self.CONFIGURATORS.get(name)))
            for name, klass in self.items()
        ]



#-*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--
"""
ElasticSearch based search engine plugin.
Depends on a running ElasticSearch cluster, somewhere.
For usage, look at the unit tests.
"""

try:
    from elasticsearch import Elasticsearch, helpers
    from elasticsearch.client import IndicesClient
except ImportError:
    es_installed = False
else:
    es_installed = True


class ESQueryMapper(object):

    def match(self, field, value):
        return {'match': {field.name: {'query': value, 'operator': 'and'}}}

    def matchany(self, schema, value):
        # _all does not make sense for fulltext searches
        # so we use a custom field for that: _full.
        # it only contains Text fields
        return {'match': {'_full': {'query': value, 'operator': 'and'}}}

    def eq(self, field, value):
        return {'term': {field.name: value}}

    def gt(self, field, value):
        return {
            'range': {
                field.name: {'gt': value}
            }
        }

    def gte(self, field, value):
        return {
            'range': {
                field.name: {'gte': value}
            }
        }

    def lt(self, field, value):
        return {
            'range': {
                field.name: {'lt': value}
            }
        }

    def lte(self, field, value):
        return {
            'range': {
                field.name: {'lte': value}
            }
        }

    def in_(self, field, value):
        return {'terms': {field.name: value}}

    def and_(self, exp1, exp2):
        return {
            'bool': {
                'must': [exp1, exp2]
            }
        }

    def or_(self, exp1, exp2):
        return {
            'bool': {
                'should': [exp1, exp2]
            }
        }


class ESSchemaMapper(object):

    FT2ES = {
        'Text': {'type': 'string',
                 'analyzer':  'autocomplete',
                 'search_analyzer': 'standard',
                 'copy_to': '_full'},
        'Keyword': {'type': 'string',
                    'index': 'not_analyzed'},
        'Attachment': {'type': 'attachment'},
        'Float': {'type': 'double'},
        'Int': {'type': 'long'},
        'Boolean': {'type': 'boolean'},
        'Datetime': {'type': 'date'}
    }

    SETTINGS = {
        "number_of_shards": 1,
        "analysis": {
            "filter": {
                "autocomplete_filter": {
                    "type":     "ngram",
                    "min_gram": 1,
                    "max_gram": 20
                }
            },
            "analyzer": {
                "autocomplete": {
                    "type":      "custom",
                    "tokenizer": "standard",
                    "filter": [
                        "lowercase",
                        "autocomplete_filter"
                    ]
                }
            }
        }
    }

    def __init__(self, idx_manager):

        self.idx_manager = idx_manager
        self.mappings = {}

    # Schema API

    def define(self, schema_name):
        properties = {'_full': {"type": "string",
                                "analyzer":  "autocomplete",
                                "search_analyzer": "standard"}}
        excludes = []
        self.mappings[schema_name] = {'properties': properties,
                                      '_source': {"excludes": excludes}}

    def define_field(self, schema_name, field_type, name, indexed, stored):
        estype = dict(self.FT2ES[field_type])
        if not indexed:
            estype['index'] = 'no'
        mapping = self.mappings[schema_name]
        mapping['properties'][name] = estype
        if not stored:
            mapping['_source']['excludes'].append(name)

    ## Specific API

    def create(self, index):
        body = {"mappings": self.mappings, "settings": self.SETTINGS}
        self.idx_manager.create(index=index, body=body)


class IndexCursor(object):

    def __init__(self, index, search_function=None):
        self.index = index
        self.es_search = search_function
        self.op = {}

    # Document API

    def insert(self, schema_name, docid, **fields):
        self._action(schema_name, docid, fields)

    def update(self, schema_name, docid, **fields):
        self._action(schema_name, docid, fields, update=True)

    # Query API

    def search(self, schema_name, fields_to_load, mapped_query, limit):
        """
        ``fields_to_load`` is  a list of field names.
        """
        dsl = mapped_query
        self.op = dict(index=self.index,
                       doc_type=schema_name,
                       body={'query': dsl},
                       size=limit)

    def get_results(self, result_factory):
        hits = self.es_search(**self.op)
        return [
            (h['_score'], result_factory(h['_id'], **h['_source']))
            for h in hits['hits']['hits']
        ]

    # Specific API

    def _action(self, schema_name, docid, fields, update=False):
        doc = 'doc' if update else '_source'
        self.op = {
            '_index': self.index,
            '_type': schema_name,
            '_op_type': 'update' if update else 'create',
            '_id': docid,
            doc: fields
        }

    def enqueue(self, queue):
        queue.append(self.op)


class ElasticSearchEngine(object):
    '''
    ElasticSearch Engine.
    '''

    # make it compatible with services
    LOAD_PRIORITY = 30

    def __init__(self, index, host=None, port=None):
        '''Only one host for now.'''
        if not es_installed:
            raise ValueError('elasticsearch not installed')

        assert(index.isalpha())
        self.init_state(index, host, port)

    def init_state(self, index, host, port):
        self._queue = []
        self.index = index
        self.host = host
        self.port = port
        if host is None:
            self.es = Elasticsearch()
        else:
            self.es = Elasticsearch(hosts=[{'host': host, 'port': port}])
        self.idx_manager = IndicesClient(self.es)
        self.mapper = ESQueryMapper()

    # be persistence friendly
    def __getstate__(self):
        return (self.index, self.host, self.port)

    def __setstate__(self, state):
        self.init_state(*state)

    def _index(self, document, update=False):
        # for efficiency, nothing is executed yet,
        # we prepare and queue the operation
        cursor = IndexCursor(self.index)
        document.save(cursor, update)
        cursor.enqueue(self._queue)

    def add_document(self, document):
        '''
        Add a document to the data store, in index (a.k.a. collection),
        under the document type.
        '''
        self._index(document)

    def delete_document(self, schema, docid):
        '''
        Remove document from index and storage.
        '''
        op = {
            '_op_type': 'delete',
            '_index': self.index,
            '_type': schema.type_name,
            '_id': docid
        }
        self._queue.append(op)

    def update_document(self, document):
        '''Update document (partial update from delta document)'''
        self._index(document, True)

    def commit(self, sync=False):
        '''
        If ``sync``, index synchronously, else let Elasticsearch
        manage its index.
        '''
        helpers.bulk(self.es, self._queue)
        if sync:
            self.idx_manager.refresh(self.index)
        self._queue = []

    def cancel(self):
        '''
        Forget operation scheduled since last commit'''
        self._queue = []

    def search(self, query, size=20):
        '''
        Search the database.
        '''
        index_cursor = IndexCursor(self.index, self.es.search)
        return query.search(index_cursor, self.mapper, size)

    def delete_collection(self):
        if self.idx_manager.exists(self.index):
            self.idx_manager.delete(index=self.index)

    def create_collection(self, schemas):
        '''
        Init the collections the first time.
        Just use once! Or you'll have to reindex all your documents.
        `schemas` is a list of Document classes or Schema instances.
        '''

        idx_manager = self.idx_manager
        if idx_manager.exists(self.index):
            idx_manager.delete(index=self.index)

        mapper = ESSchemaMapper(idx_manager)
        for schema in schemas:
            schema.map(mapper)

        mapper.create(self.index)



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import presentation, security

from .comp import CardMembers


@presentation.render_for(CardMembers, 'action')
def render_card_members(self, h, comp, model):
    """Member section view for card

    First members icons,
    Then icon "more user" if necessary
    And at the end icon "add user"
    """
    if not security.has_permissions('edit', self.card):
        return h.root
    with h.div(class_='members'):
        h << h.script('''YAHOO.kansha.app.hideOverlay();''')
        for m in self.members[:self.MAX_SHOWN_MEMBERS]:
            h << m.on_answer(self.remove_member).render(h, model="overlay-remove")
        if len(self.members) > self.MAX_SHOWN_MEMBERS:
            h << h.div(self.see_all_members, class_='more')
        h << h.div(self.overlay_add_members, class_='add')
    return h.root


@presentation.render_for(CardMembers, 'badge')
@presentation.render_for(CardMembers, model='members_read_only')
def render_card_members_read_only(self, h, comp, model):
    """Member section view for card

    First members icons,
    Then icon "more user" if necessary
    And at the end icon "add user"
    """
    with h.div(class_='members'):
        for m in self.members[:self.MAX_SHOWN_MEMBERS]:
            h << h.span(m.render(h, 'avatar'), class_='miniavatar unselectable')
        if len(self.members) > self.MAX_SHOWN_MEMBERS:
            if model == 'badge':
                h << comp.render(h, 'more_users')
            else:
                h << h.div(self.see_all_members, class_='more')
    return h.root


@presentation.render_for(CardMembers, 'members_list_overlay')
def render_members_members_list_overlay(self, h, comp, *args):
    """Overlay to list all members"""
    h << h.h2(_('All members'))
    # with h.form:
    with h.div(class_="members"):
        if security.has_permissions('edit', self.card):
            h << [m.on_answer(comp.answer).render(h, "remove") for m in self.members]
        else:
            h << [m.render(h, "avatar") for m in self.members]
    return h.root


def _add_members(member_ext, members):
    member_ext.add_members(members)
    return "YAHOO.kansha.reload_cards['%s']();YAHOO.kansha.app.hideOverlay();" % member_ext.card.id


@presentation.render_for(CardMembers, 'add_member_overlay')
def render_members_add_member_overlay(self, h, comp, *args):
    """Overlay to add member"""
    h << h.h2(_('Add members'))
    favorites = self.favorites
    if favorites:
        with h.div(class_='favorites'):
            h << h.h3(_('Suggestions'))
            with h.ul:
                for favorite in favorites:
                    h << h.li(favorite.on_answer(lambda members: _add_members(self, members)))
    with h.div(class_='members search'):
        h << self.new_member.on_answer(lambda members: _add_members(self, members))
    return h.root


@presentation.render_for(CardMembers, 'more_users')
def render_members_many_user(self, h, comp, *args):
    number = len(self.members) - self.MAX_SHOWN_MEMBERS
    return h.span(
        h.i(class_='ico-btn icon-user'),
        h.span(number, class_='count'),
        title=_('%s more...') % number)



"""add display week numbers

Revision ID: 1bd634091036
Revises: 4fb34b202c5f
Create Date: 2015-09-15 11:51:46.739150

"""

# revision identifiers, used by Alembic.
revision = '1bd634091036'
down_revision = '4fb34b202c5f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('user', sa.Column('display_week_numbers', sa.Boolean))


def downgrade():
    op.drop_column('user', 'display_week_numbers')



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from elixir import using_options
from elixir import ManyToOne
from elixir import Field, UnicodeText, DateTime
from sqlalchemy import func

from kansha.models import Entity


class DataComment(Entity):

    """Comment mapper
    """
    using_options(tablename='comment')
    comment = Field(UnicodeText, default=u'')
    creation_date = Field(DateTime)
    card = ManyToOne('DataCard')
    author = ManyToOne('DataUser')

    @classmethod
    def get_by_card(cls, card):
        q = cls.query
        q = q.filter_by(card=card)
        q = q.order_by(cls.creation_date.desc())
        return q

    @classmethod
    def total_comments(cls, card):
        q = cls.query
        q = q.filter_by(card=card)
        # query.count() is sloooow, so we use an alternate method
        return q.with_entities(func.count()).scalar()



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import presentation, var

from .comp import EditableTitle, Title


@presentation.render_for(Title)
def render_title(self, h, comp, model):
    title = self.title()
    if title is not None and len(title) < 5:
        title += u' ' * (5 - len(title)) # unbreakable spaces ; when display-block is NOT an option
    with h.a(class_=self.class_, title=title).action(comp.answer):
        h << (h.input(placeholder=self.placeholder) if (self.placeholder and title is None) else title)

    return h.root


@presentation.render_for(Title, 'readonly')
def render_title_ro(self, h, *args):
    return h.span(self.title() or '', class_=self.class_)


@presentation.render_for(Title, 'edit')
def render_title_edit(self, h, comp, *args):
    title = var.Var(self.title() or '')
    id_ = h.generate_id('id')

    with h.form(class_=self.class_ + '-form'):
        input = (h.textarea(title, style='height: %dpx' % self.height) if self.height else h.input(value=title))
        h << input(id=id_, placeholder=self.placeholder).action(title)

        h << h.button(_('Save'), class_='btn btn-primary').action(self.set_title, comp, title)
        h << ' '
        h << h.button(_('Cancel'), class_='btn').action(comp.answer)

        h << h.script('YAHOO.kansha.app.selectElement(%s);' % id_)

    return h.root


@presentation.render_for(EditableTitle, 'readonly')
def render_editabletitle_ro(self, h, *args):
    return self.title.render(h, 'readonly')



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --


class EventHandlerMixIn(object):
    """
    Mix-in that implements:
      - `emit_event`, to emit an event;
      - `handle_event`, a callback for comp.on_answer if comp is expected to emit events.

      `handle_event` calls a method `on_event(event)`
      on `self` (if exists) to handle the event and then systematically bubbles the event up.
      `handle_event` returns the return value of `on_event`
      if any, or the return value of the upper levels.
    """

    def emit_event(self, comp, kind, data=None):
        event = kind(data, source=[self])
        return comp.answer(event)

    def handle_event(self, comp, event):

        local_res = None
        local_handler = getattr(self, 'on_event', None)
        if local_handler:
            local_res = local_handler(comp, event)
        # bubble up in any case
        event.append(self)
        upper_res = comp.answer(event)
        # return local result in priority
        return local_res or upper_res


class Event(object):
    """Can be derived"""

    def __init__(self, data, source=[]):
        """
        `data` is a payload specific to the kind of event.
        `source` is a list of component business objects. The first one is the emitter.
        Each traversed component must append itself with `append` (see below).
        """

        self._source = source
        self.data = data

    @property
    def source(self):
        return self._source.copy()

    @property
    def emitter(self):
        return self._source[0]

    @property
    def last_relay(self):
        return self._source[-1]

    def is_(self, kind):
        return type(self) is kind

    def is_kind_of(self, kind):
        return isinstance(self, kind)

    def append(self, relay):
        self._source.append(relay)

    def cast_as(self, sub_kind):
        #assert(issubclass(sub_kind, self.__class__))
        return sub_kind(self.data, self._source)


# Standard events

class ColumnDeleted(Event):
    """
    The user clicked on the 'delete column' action.
    `data` is the column component (component.Component).
    """
    pass


class CardClicked(Event):
    """
    The user clicked on a card.
    `data` is the card component (component.Component)
    """
    pass


class PopinClosed(Event):
    """
    The Popin has closed.
    `data` is the component.Component containing the Popin.
    """
    pass


class CardEditorClosed(PopinClosed):
    """
    In the particular case when the Popin contains the card editor.
    `data` is the component.Component containing the Popin.
    """
    pass


class CardArchived(Event):
    """
    The user clicked on the `Delete` button in the card editor.
    No payload.
    """
    pass


class SearchIndexUpdated(Event):
    """
    Some operations have been committed on the search index.
    No payload.
    """
    pass


class CardDisplayed(Event):
    """
    A card has just been (re-)displayed on the board (default form).
    No payload.
    """
    pass


class BoardAccessChanged(Event):
    """
    The access conditions to the board changed.
    """


class BoardDeleted(BoardAccessChanged):
    """
    The board has been (or is about to be) deleted.
    No payload.
    """


class BoardArchived(BoardAccessChanged):
    """
    The board has been archived.
    No payload.
    """


class BoardRestored(BoardAccessChanged):
    """
    The board has been restored from archive.
    """


class BoardLeft(BoardAccessChanged):
    """
    The user has left the board.
    No payload.
    """


# "Request" events
class ParentTitleNeeded(Event):
    """The emitter needs context from parent in the form of a title string."""
    pass


class NewTemplateRequested(Event):
    """
    The user requested that a new template is created from the emitter.
    Payload is tuple (template_title, template_description, shared_flag).
    The receiver returns a new Template on success.
    """
    pass



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from kansha.cardextension.tests import CardExtensionTestCase

from .comp import Checklists

class ChecklistTest(CardExtensionTestCase):

    extension_name = 'checklists'
    extension_class = Checklists

    def test_add_delete(self):
        self.assertEqual(len(self.extension.checklists), 0)
        self.extension.add_checklist()
        self.assertEqual(len(self.extension.checklists), 1)
        with self.assertRaises(IndexError):
            self.extension.delete_checklist(1)
        self.extension.delete_checklist(0)
        self.assertEqual(len(self.extension.checklists), 0)

    def test_title(self):
        ck = self.extension.add_checklist()
        ck.set_title(u'test')
        self.assertEqual(ck.data.title, u'test')
        self.assertEqual(ck.get_title(), u'test')

    def test_items(self):
        ck = self.extension.add_checklist()
        self.assertEqual(ck.total_items, 0)
        ck.add_item_from_str(u'test')
        self.assertEqual(ck.total_items, 1)
        ck.add_item_from_str(u'test2')
        self.assertEqual(ck.total_items, 2)
        ck.add_item_from_str(u'test3')
        self.assertEqual(ck.total_items, 3)
        ck.delete_index(1)
        self.assertEqual(ck.total_items, 2)
        ck.set_index(1)
        self.assertEqual(ck.data.index, 1)
        self.assertEqual(ck.items[0]().get_title(), u'test')
        self.assertEqual(ck.items[1]().get_title(), u'test3')

    def test_item(self):
        ck = self.extension.add_checklist()
        self.assertEqual(ck.total_items, 0)
        self.assertEqual(ck.total_items_done, 0)
        self.assertEqual(ck.progress, 0)
        ck.add_item_from_str(u'test')
        self.assertEqual(ck.total_items, 1)
        self.assertEqual(ck.total_items_done, 0)
        self.assertEqual(ck.progress, 0)
        item = ck.items[0]()
        self.assertEqual(item.get_title(), u'test')
        item.set_title(u'test2')
        self.assertEqual(item.get_title(), u'test2')
        self.assertFalse(item.done)
        item.set_done()
        self.assertTrue(item.done)
        self.assertEqual(ck.total_items, 1)
        self.assertEqual(ck.total_items_done, 1)
        self.assertEqual(ck.progress, 100)
        ck.add_item_from_str(u'test')
        self.assertEqual(ck.progress, 50)
        item.set_done()
        self.assertFalse(item.done)

    def test_copy(self):
        ck = self.extension.add_checklist()
        ck.set_title(u'test')
        ck.set_index(0)
        ck.add_item_from_str(u'test')
        ck.add_item_from_str(u'test2')
        ck.items[0]().set_title(u'test item')
        item = ck.items[1]()
        item.set_title(u'test item2')
        item.set_done()
        ck = self.extension.add_checklist()
        ck.set_title(u'test2')
        ck.set_index(1)
        ck.add_item_from_str(u'test3')

        cpy = self.extension_copy
        self.assertEqual(len(cpy.checklists), 2)
        self.assertEqual(cpy.total_items, 3)
        self.assertEqual(cpy.total_items_done, 0)
        ck = cpy.checklists[0]()
        self.assertEqual(ck.get_title(), u'test')
        self.assertEqual(ck.data.index, 0)
        item = ck.items[0]()
        self.assertEqual(item.get_title(), u'test item')

    def test_update_document(self):
        doc = self.card.schema(docid=None)
        ck = self.extension.add_checklist()
        ck.set_title(u'test list')
        ck.add_item_from_str(u'test item')
        ck.add_item_from_str(u'test item 2')

        ck = self.extension.add_checklist()
        ck.set_title(u'test list 2')
        ck.add_item_from_str(u'test item 3')
        ck.add_item_from_str(u'test item 4')
        self.extension.update_document(doc)
        self.assertEqual(doc.checklists, u'''test list
test item
test item 2
test list 2
test item 3
test item 4''')





"""empty message

Revision ID: 2c0a3788227
Revises: f058ce7ee0c
Create Date: 2016-01-20 17:18:51.047928

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = '2c0a3788227'
down_revision = 'f058ce7ee0c'


def upgrade():
    bind = op.get_bind()
    if bind.engine.name != 'sqlite':
        op.drop_column('user', 'display_week_numbers')


def downgrade():
    op.add_column('user', sa.Column('display_week_numbers', sa.Boolean))



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

"""adding max cards

Revision ID: b740362087
Revises: 537fa16b46e7
Create Date: 2013-09-19 17:37:37.027495

"""

# revision identifiers, used by Alembic.
revision = 'b740362087'
down_revision = '537fa16b46e7'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('column', sa.Column('nb_max_cards', sa.Integer))


def downgrade():
    op.drop_column('column', 'nb_max_cards')



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

import peak
import datetime

from nagare import presentation, security, ajax, i18n
from nagare.i18n import _, format_date

from .comp import DueDate


@peak.rules.when(ajax.py2js, (datetime.date,))
def py2js(value, h):
    """Generic method to transcode a Datetime

    In:
      - ``value`` -- the datetime object
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    dt = i18n.to_timezone(value)
    return 'new Date("%s", "%s", "%s")' % (
        dt.year, dt.month - 1, dt.day)


@peak.rules.when(ajax.py2js, (DueDate,))
def py2js(value, h):
    if value.due_date:
        return ajax.py2js(value.due_date, h)
    return None


@presentation.render_for(DueDate)
def render_DueDate(self, h, comp, model):
    return h.root


@presentation.render_for(DueDate, model='badge')
def render_DueDate_badge(self, h, *args):
    """Gallery badge for the card"""
    if self.due_date:
        with h.span(class_='badge'):
            h << h.span(h.i(class_='icon-alarm'), ' ', self.get_days_count(), class_='label due-date ' + self.get_class(), title=format_date(self.due_date, 'full'))
    return h.root


@presentation.render_for(DueDate, model='action')
def render_DueDate_button(self, h, comp, *args):
    if security.has_permissions('due_date', self.card):
        id_ = h.generate_id()
        if self.due_date:
            classes = ['btn', 'btn-due-date', self.get_class()]
            with h.a(class_=u' '.join(classes), id_=id_).action(self.calendar().toggle):
                h << h.i(class_='icon-alarm duedate-icon')
                h << format_date(self.due_date, 'short')
        else:
            with h.a(class_='btn', id_=id_).action(self.calendar().toggle):
                h << h.i(class_='icon-alarm')
                h << _('Due date')
        h << self.calendar.on_answer(self.set_value)
    return h.root



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import presentation, component, ajax, var

from kansha import validator
from kansha.toolbox import overlay, remote

from .usermanager import NewMember
from .comp import User, PendingUser


@presentation.render_for(User)
def render_User(self, h, comp, *args):
    """Default view, render avatar and fullname"""
    h << comp.render(h, "avatar")
    with h.div(class_="name"):
        h << comp.render(h, model="fullname")
        h << h.span(self.data.email or self.data.email_to_confirm, class_="email")
    return h.root


@presentation.render_for(User, model='avatar')
def render_User_avatar(self, h, comp, model, *args):
    """Render a user's avatar"""
    with h.span(class_='avatar', title='%s' % (self.data.fullname)):
        avatar = self.get_avatar()
        if avatar:
            h << h.img(src=avatar)
        else:
            h << h.i(class_='ico-btn icon-user')
    return h.root


@presentation.render_for(User, model='fullname')
def render_User_fullname(self, h, comp, model, *args):
    """Render a user's avatar"""
    return h.span(self.data.fullname, class_="fullname")


@presentation.render_for(User, model='search')
def render_User_search(self, h, comp, *args):
    """Serach result view"""
    return h.div(comp)


@presentation.render_for(User, model='overlay-member')
@presentation.render_for(User, model='overlay-manager')
@presentation.render_for(User, model='overlay-remove')
@presentation.render_for(User, model='overlay-last_manager')
@presentation.render_for(PendingUser, model='overlay-pending')
def render_User_overlay_remove(self, h, comp, model):
    sub_model = model[8:]
    h << component.Component(
        overlay.Overlay(
            lambda r: (
                comp.render(r, "avatar"),
                {'class': 'miniavatar'}
            ),
            lambda r: comp.render(r, sub_model),
            dynamic=False,
            cls='card-overlay' if model == 'remove' else 'board-labels-overlay'
        )
    )
    return h.root


@presentation.render_for(User, "remove")
def render_User_remove(self, h, comp, *args):
    """Render for search result"""
    with h.div(class_='member'):
        h << comp.render(h, None)
        with h.span(class_="actions"):
            with h.form:
                h << h.input(value=_("Remove"), type="submit",
                             class_="btn btn-primary delete").action(ajax.Update(action=lambda: comp.answer(self.username)))
    return h.root


@presentation.render_for(User, "member")
@presentation.render_for(User, "manager")
def render_User_manager(self, h, comp, model):
    """Render for search result"""
    with h.div(class_='member'):
        h << comp.render(h, None)
        with h.span(class_="actions"):
            with h.form:
                h << h.input(
                    value=_("Manager"),
                    type="submit",
                    class_=("btn btn-primary toggle" if model == 'manager'
                            else "btn btn-primary")
                ).action(ajax.Update(action=lambda: comp.answer('toggle_role')))
                h << h.input(
                    value=_("Remove"),
                    type="submit",
                    class_="btn btn-primary delete"
                ).action(ajax.Update(action=lambda: comp.answer('remove')))
    return h.root


@presentation.render_for(User, "last_manager")
def render_User_last_manager(self, h, comp, *args):
    """Render for search result"""
    with h.div(class_='member'):
        h << comp.render(h, None)
        with h.span(class_="actions last-manager"):
            h << _("You are the last manager, you can't leave this board.")
    return h.root


@presentation.render_for(User, model="menu")
def render_User_menu(self, h, comp, *args):
    """Render user menu"""
    h << h.li(h.a(h.i(class_='icon-home'), _("Home"), class_="home").action(lambda: comp.answer(self)))
    h << h.li(h.a(h.i(class_='icon-switch'), _("Logout"), class_="logout").action(comp.answer))
    return h.root


@presentation.render_for(User, model="detailed")
def render_User_detailed(self, h, comp, *args):
    """Render user detailed view"""
    h << h.h3(self.data.fullname)
    h << h.div(comp.render(h, model='avatar'))
    return h.root


@presentation.render_for(User, model="friend")
def render_User_friend(self, h, comp, *args):
    h << h.a(comp.render(h, "avatar")).action(
        remote.Action(lambda: comp.answer([self.data.email or self.data.email_to_confirm])))
    return h.root


@presentation.render_for(NewMember, model='add_members')
def render_AddMembers(self, h, comp, *args):
    value = var.Var('')
    submit_id = h.generate_id('form')
    hidden_id = h.generate_id('hidden')

    with h.form:
        h << h.input(type='text', id=self.text_id)
        h << h.input(type='hidden', id=hidden_id).action(value)
        h << self.autocomplete
        h << h.script(
            u"%(ac_id)s.itemSelectEvent.subscribe(function(sType, aArgs) {"
            u"var value = aArgs[2][0];"
            u"YAHOO.util.Dom.setAttribute(%(hidden_id)s, 'value', value);"
            u"YAHOO.util.Dom.get(%(submit_id)s).click();"
            u"});" % {
                'ac_id': self.autocomplete().var,
                'hidden_id': ajax.py2js(hidden_id),
                'submit_id': ajax.py2js(submit_id)
            }
        )
        h << h.script(
            "document.getElementById(%s).focus()" % ajax.py2js(self.text_id)
        )
        h << h.button(id=submit_id, style='display:none').action(remote.Action(lambda: comp.answer([] if not value() else [value()])))
    return h.root


@presentation.render_for(NewMember)
def render_NewMember(self, h, comp, *args):
    """Render the title of the associated card"""
    with h.form:
        members_to_add = var.Var()

        def get_emails():
            emails = []
            for email in members_to_add().split(','):
                try:
                    email = email.strip()
                    email = validator.validate_email(email)
                    emails.append(email)
                except ValueError:
                    continue
            return emails

        h << h.input(type='text', id=self.text_id,).action(members_to_add)
        mail_input = h.input(value=_("Add"), type="submit",
                             class_="btn btn-primary"
                            ).action(remote.Action(lambda: comp.answer(get_emails())))
        # Sending mail synchronously can take a long time
        mail_input.set('onclick', 'YAHOO.kansha.app.showWaiter();' + mail_input.get('onclick'))
        h << mail_input
        h << self.autocomplete
    h << h.script(
        "document.getElementById(%s).focus()" % ajax.py2js(self.text_id)
    )
    return h.root

# Pending User


@presentation.render_for(PendingUser)
def render_PendingUser(self, h, comp, *args):
    """Default view, render avatar and fullname"""
    h << comp.render(h, "avatar")
    with h.div(class_="name"):
        h << h.span(self.username, class_="fullname")
        h << h.span(self.username, class_="email")
    return h.root


@presentation.render_for(PendingUser, model='avatar')
def render_PendingUser_avatar(self, h, comp, model, *args):
    """Render a user's avatar"""
    with h.span(class_='avatar', title='%s' % (self.username)):
        h << h.i(class_='ico-btn icon-mail2')
    return h.root


@presentation.render_for(PendingUser, "pending")
def render_PendingUser_pending(self, h, comp, *args):
    """Render for search result"""
    with h.div(class_='member'):
        h << comp.render(h, None)
        with h.span(class_="actions"):
            with h.form:
                h << h.input(value=_("Resend invitation"), type="submit",
                             class_="btn btn-primary").action(ajax.Update(action=lambda: comp.answer('resend')))
                h << h.input(value=_("Remove"), type="submit",
                             class_="btn btn-primary delete").action(ajax.Update(action=(lambda: comp.answer('remove'))))
    return h.root



#-*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--
"""
Schemas are used to define Documents.
Example usage:

class MyDoc(Document):
    title = Text(stored=True)
    created = Datetime
    description = Text

doc = MyDoc(1001, title=u'Programming Python')
The first argument is a mandatory, arbitrary id.

"""

# TODO
# type checking, linguistic settings (language, collations, stop words,
# lexers, ...)

from datetime import datetime
from .query import *


class FieldType(object):

    schema = None

    def __init__(self, name='', indexed=True, stored=False, default=None):
        self.name = name
        self.indexed = indexed
        self.stored = stored
        if default is not None:
            self.default = default

    @property
    def type_name(self):
        return self.__class__.__name__

    def __eq__(self, v):
        return EQQuery(self, v)

    def __ne__(self, v):
        return NEQuery(self, v)

    def __gt__(self, v):
        return GTQuery(self, v)

    def __ge__(self, v):
        return GTEQuery(self, v)

    def __lt__(self, v):
        return LTQuery(self, v)

    def __le__(self, v):
        return LTEQuery(self, v)

    def in_(self, v):
        return INQuery(self, v)

    def match(self, v):
        return MATCHQuery(self, v)

    def match_phrase(self, v):
        return PHRASEQuery(self, v)

    def map(self, schema_name, mapper):
        mapper.define_field(
            schema_name,
            self.type_name,
            self.name,
            self.indexed,
            self.stored
        )


class Text(FieldType):

    '''Fulltext field'''
    default = u''


class Keyword(FieldType):

    '''Exact match string'''
    default = u''


class Attachment(FieldType):

    '''
    File, as a file path unicode string, whose content
    is to be indexed.
    '''
    default = None


class Number(FieldType):
    pass


class Float(Number):
    default = 0.0


class Int(Number):
    default = 0


class Boolean(FieldType):
    default = True


class Datetime(FieldType):
    default = datetime.now()


class IndexableDocument(object):
    '''Document Type for documents instanciated by a Schema object.'''

    def __init__(self, schema, fields, docid, **field_values):
        self._id = docid
        self.fields = fields
        self.schema = schema
        for name, field in fields.iteritems():
            setattr(self, name, field_values.get(name, field.default))

    @property
    def schema_name(self):
        return self.schema.type_name

    def save(self, cursor, update=False):
        fields = dict(
            (name, getattr(self, name)) for name in self.fields
            if getattr(self, name) is not None
        )
        if update:
            cursor.update(self.schema_name, docid=self._id, **fields)
        else:
            cursor.insert(self.schema_name, docid=self._id, **fields)


class _Schema(type):
    """
    Meta-class
    """

    def __new__(cls, name, bases, dct):
        fields = {}
        for aname, aval in dct.items():
            if isinstance(aval, FieldType):
                aval.name = aname
                fields[aname] = aval
            elif isinstance(aval, type) and issubclass(aval, FieldType):
                val = aval()
                val.name = aname
                fields[aname] = dct[aname] = val
        klass = super(_Schema, cls).__new__(cls, name, bases, dct)
        klass.fields = fields
        klass.type_name = name
        for v in fields.itervalues():
            v.schema = klass
        return klass


class Document(IndexableDocument):

    '''
    Declarative class for documents.

    The `fields`, 'delta' , 'type_name', 'schema' and `match` attributes are reserved: don't declare properties with those names!

    Usage:

        class MyDocument(schema.Document):
            title = schema.Text(stored=True)
            tags = schema.Text(stored=False)
            pages = schema.Int(stored=True)
            description = schema.Text
            price = schema.Float(stored=True, indexed=False)

        doc1 = MyDocument(
                    'doc1', title=u'Titre', tags=u'Un livre français best seller',
                    pages=89, description=u'Description de qualité, avec des services.', price=10.0)
        query = MyDocument.match(u'tests')
        query = MyDocument.tags.match(u'best')
        query = (MyDocument.pages == 89) | MyDocument.description.match(
                    u'practices')

    '''

    __metaclass__ = _Schema

    type_name = 'Document'

    def __init__(self, docid, **fields):
        '''
        Instanciate new document.
        The caller is responsible for giving a (preferably) unique id to the document.
        The id is not part of the schema declaration and is mandatory.
        it can be used to track the document's origin for example.
        Fields values not specified to the constructor are set to their defaults.
        Fields set to None are ignored (that is, in case of update,
        they are left untouched in the index).
        '''
        self._id = docid
        for fname, fvalue in self.fields.iteritems():
            setattr(self, fname, fields.get(fname, fvalue.default))

    @property
    def schema(self):
        return self.__class__

    @property
    def schema_name(self):
        return self.__class__.type_name

    @classmethod
    def delta(cls, docid, **fields):
        '''
        Alternate constructor to build delta documents.
        Fields values not specified to the constructor are set to None.
        '''
        for fname in set(cls.fields) - set(fields):
            fields[fname] = None
        return cls(docid, **fields)

    @classmethod
    def match(cls, v):
        '''Full text specific: match any of the fields'''
        return MATCHANYQuery(cls, v)

    @classmethod
    def map(cls, mapper):
        mapper.define(cls.type_name)
        for field in cls.fields.itervalues():
            field.map(cls.type_name, mapper)

    @classmethod
    def iter_fields(cls):
        return cls.fields.iteritems()


class Schema(object):
    '''Imperatively build a Document schema.

    Same reserved keywords as the Declarative Schema.

    Usage:

        TestDocument = schema.Schema(
            'MyDocument',
            schema.Text('title', stored=True),
            schema.Text('tags', stored=False),
            schema.Int('pages', stored=True),
            schema.Text('description'),
            schema.Float('price', stored=True, indexed=False)
        )

    or

        # Incrementaly update schema "in place"
        TestDocument = schema.Schema('MyDocument')
        TestDocument.add_field(schema.Text('title', stored=True))
        TestDocument.add_field(schema.Text('tags', stored=False))
        TestDocument.add_field(schema.Int('pages', stored=True))
        TestDocument.add_field(schema.Text('description'))
        TestDocument.add_field(schema.Float('price', stored=True, indexed=False))

    or create new schema from existing one

        TestDocument2 = TestDocument + schema.Text('author')
        TestDocument2.type_name = 'IdentifiedDocument'

    TestDocument is then used as if it was a Declarative Document schema.
    '''

    def __init__(self, name, *fields):
        self.type_name = name
        self.fields = {}
        for field in fields:
            self.add_field(field)

    def add_field(self, field):
        """Update schema in place."""
        assert(field.name)
        assert(field.name != 'schema')
        assert(not hasattr(self, field.name))
        field.schema = self
        self.fields[field.name] = field

    def __add__(self, field):
        """Create a new augmented schema."""
        new_schema = Schema(self.type_name, *self.fields.values())
        new_schema.add_field(field)
        return new_schema

    def __call__(self, docid, **fields):
        '''
        Document factory.
        Instanciate new document.
        The caller is responsible for giving a (preferably) unique id to the document.
        The id is not part of the schema declaration and is mandatory.
        it can be used to track the document's origin for example.
        Fields values not specified to the constructor are set to their defaults.
        Fields set to None are ignored (that is, in case of update,
        they are left untouched in the index).
        '''
        return IndexableDocument(self, self.fields, docid, **fields)

    def delta(self, docid, **fields):
        '''
        Alternate constructor to build delta documents.
        Fields values not specified to the constructor are set to None.
        '''
        for fname in set(self.fields) - set(fields):
            fields[fname] = None
        return self(docid, **fields)

    def match(self, v):
        '''Full text specific: match any of the fields'''
        return MATCHANYQuery(self, v)

    def __getattr__(self, name):
        try:
            return self.fields[name]
        except KeyError:
            raise AttributeError(name)

    def __getstate__(self):
        return (self.type_name, self.fields)

    def __setstate__(self, data):
        self.type_name, self.fields = data

    def map(self, mapper):
        mapper.define(self.type_name)
        for field in self.fields.itervalues():
            field.map(self.type_name, mapper)

    def iter_fields(self):
        return self.fields.iteritems()



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

import datetime

from peak.rules import when

from nagare.i18n import _
from nagare.database import session
from nagare.security import common
from nagare import component, security

from kansha import validator
from kansha.card import Card
from kansha.user import usermanager
from kansha.board import excel_export
from kansha.services.search import schema
from kansha.cardextension import CardExtension
from kansha.board import COMMENTS_MEMBERS, COMMENTS_PUBLIC
from kansha.services.actionlog.messages import render_event

from .models import DataComment


@when(common.Rules.has_permission, "user and (perm == 'delete_comment')")
def has_permission_delete_comment(self, user, perm, comment):
    return comment.is_author(user)


@when(common.Rules.has_permission, "user and (perm == 'edit_comment')")
def has_permission_edit_comment(self, user, perm, comment):
    return comment.is_author(user)


@when(render_event, "action=='card_add_comment'")
def render_event_card_add_comment(action, data):
    return _(u'User %(author)s has commented card "%(card)s"') % data


class Comment(object):

    """Comment component"""

    def __init__(self, data):
        """Initialization

        In:
            - ``data`` -- the comment object from the database
        """
        self.db_id = data.id
        self.text = data.comment
        self.creation_date = data.creation_date
        self.author = component.Component(usermanager.UserManager.get_app_user(data.author.username, data=data.author))
        self.comment_label = component.Component(Commentlabel(self))
        self.comment_label.on_answer(lambda v: self.comment_label.call(model='edit'))

    def edit_comment(self):
        self.comment_label.call(model='edit')

    def is_author(self, user):
        return self.author().username == user.username

    def set_comment(self, text):
        if text is None:
            return
        text = text.strip()

        if text:
            self.data.comment = validator.clean_text(text)

    @property
    def data(self):
        """Return the comment object from the database
        """
        return DataComment.get(self.db_id)


class Commentlabel(object):

    """comment label component.

    """

    def __init__(self, parent):
        """Initialization

        In:
          - ``parent`` -- parent of comment label
        """
        self.parent = parent
        self.text = parent.text or u''

    def change_text(self, text):
        """Change the comment of our wrapped object

        In:
            - ``text`` -- the new comment
        Return:
            - the new comment

        """
        if text is None:
            return
        text = text.strip()
        if text:
            text = validator.clean_text(text)
            self.text = text
            self.parent.set_comment(text)
            return text

    def is_author(self, user):
        return self.parent.is_author(user)


class Comments(CardExtension):

    LOAD_PRIORITY = 50

    """Comments component
    """

    def __init__(self, card, action_log, configurator):
        """Initialization

        In:
            - ``parent`` -- the parent card
            - ``comments`` -- the comments of the card
        """
        super(Comments, self).__init__(card, action_log, configurator)
        self.comments = []

    def load_children(self):
        if not self.comments:
            self.comments = [self._create_comment_component(data_comment) for data_comment in self.data]

    @staticmethod
    def get_schema_def():
        return schema.Text(u'comments')

    def update_document(self, document):
        self.load_children()
        document.comments = u'\n'.join(comment().text for comment in self.comments)

    @property
    def data(self):
        return DataComment.get_by_card(self.card.data)

    @property
    def allowed(self):
        return self.configurator.comments_allowed

    def _create_comment_component(self, data_comment):
        return component.Component(Comment(data_comment)).on_answer(self.delete_comment)

    def add(self, v):
        """Add a new comment to the card

        In:
            - ``v`` -- the comment content
        """
        security.check_permissions('comment', self)
        if v is None:
            return
        v = v.strip()
        if v:
            v = validator.clean_text(v)
            user = security.get_user()
            comment = DataComment(comment=v, card_id=self.card.db_id,
                                  author=user.data, creation_date=datetime.datetime.utcnow())
            session.add(comment)
            session.flush()
            data = {'comment': v.strip(), 'card': self.card.get_title()}
            self.action_log.add_history(security.get_user(), u'card_add_comment', data)
            self.comments.insert(0, self._create_comment_component(comment))

    def delete_comment(self, comp):
        """Delete a comment.

        In:
            - ``comment`` -- the comment to delete
        """
        self.comments.remove(comp)
        comment = comp()
        DataComment.get(comment.db_id).delete()
        session.flush()

    def delete(self):
        self.comments = []
        for comment in self.data:
            comment.delete()

    @property
    def num_comments(self):
        return DataComment.total_comments(self.card.data)


@excel_export.get_extension_title_for(Comments)
def get_extension_title_Comments(card_extension):
    return _(u'Comments')


@excel_export.write_extension_data_for(Comments)
def write_extension_data_Comments(self, sheet, row, col, style):
    self.load_children()
    comments = u'\n-----\n'.join(comment().text for comment in self.comments)
    sheet.write(row, col, comments, style)


# TODO: completely redefine security
@when(common.Rules.has_permission, "user and perm == 'comment' and isinstance(subject, Comments)")
def has_permission_Card_comment(self, user, perm, comments):
    return ((security.has_permissions('edit', comments.card) and comments.allowed == COMMENTS_MEMBERS) or
            (comments.allowed == COMMENTS_PUBLIC and user))



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import json
import itertools
from functools import partial

from nagare.i18n import _
from peak.rules import when
from nagare.security import common
from nagare.database import session
from nagare import component, log, security, var

from kansha import title
from kansha.card import Card
from kansha.user import usermanager
from kansha.services import ActionLog
from kansha.column import comp as column
from kansha.user.comp import PendingUser
from kansha.toolbox import popin, overlay
from kansha.card_addons.label import Label
from kansha.authentication.database import forms
from kansha import events, exceptions, validator
from kansha.board_card_filter import BoardCardFilter

from .boardconfig import BoardConfig
from .excel_export import ExcelExport
from .templates import SaveTemplateTask
from .models import DataBoard, BOARD_PRIVATE, BOARD_PUBLIC, BOARD_SHARED


# Votes authorizations
VOTES_OFF = 0
VOTES_MEMBERS = 1
VOTES_PUBLIC = 2

# Comments authorizations
COMMENTS_OFF = 0
COMMENTS_MEMBERS = 1
COMMENTS_PUBLIC = 2


# WEIGHTING CARDS
WEIGHTING_OFF = 0
WEIGHTING_FREE = 1
WEIGHTING_LIST = 2


class Board(events.EventHandlerMixIn):

    """Board component"""

    MAX_SHOWN_MEMBERS = 4
    background_max_size = 3 * 1024  # in Bytes

    def __init__(self, id_, app_title, app_banner, theme, card_extensions, search_engine_service,
                 assets_manager_service, mail_sender_service, services_service,
                 load_children=True, data=None):
        """Initialization

        In:
          -- ``id_`` -- the id of the board in the database
          -- ``mail_sender_service`` -- Mail service, used to send mail
          -- ``on_board_delete`` -- function to call when the board is deleted
        """
        self.model = 'columns'
        self.app_title = app_title
        self.app_banner = app_banner
        self.theme = theme
        self.mail_sender = mail_sender_service
        self.id = id_
        self._data = data
        self.assets_manager = assets_manager_service
        self.search_engine = search_engine_service
        self._services = services_service
        # Board extensions are not extracted yet, so
        # board itself implement their API.
        self.board_extensions = {
            'weight': self,
            'labels': self,
            'members': self,
            'comments': self,
            'votes': self
        }
        self.card_extensions = card_extensions.set_configurators(self.board_extensions)

        self.action_log = ActionLog(self)

        self.version = self.data.version
        self.modal = component.Component(popin.Empty())
        self.card_filter = self._services(BoardCardFilter, Card.schema, self.id,
                                          not self.show_archive)
        self.search_input = component.Component(self.card_filter, 'search_input')

        self.columns = []
        self.archive_column = None
        if load_children:
            self.load_children()

        # Member part
        self.overlay_add_members = component.Component(
            overlay.Overlay(lambda r: (r.i(class_='ico-btn icon-user-plus')),
                            lambda r: component.Component(self).render(r, model='add_member_overlay'),
                            dynamic=True, cls='board-labels-overlay'))
        self.new_member = component.Component(usermanager.NewMember(self.autocomplete_method))

        self.update_members()

        def many_user_render(h, number):
            return h.span(
                h.i(class_='ico-btn icon-user'),
                h.span(number, class_='count'),
                title=_("%s more...") % number)

        self.see_all_members = component.Component(overlay.Overlay(lambda r: many_user_render(r, len(self.all_members) - self.MAX_SHOWN_MEMBERS),
                                                                   lambda r: component.Component(self).render(r, model='members_list_overlay'),
                                                                   dynamic=False, cls='board-labels-overlay'))
        self.see_all_members_compact = component.Component(overlay.Overlay(lambda r: many_user_render(r, len(self.all_members)),
                                                                           lambda r: component.Component(self).render(r, model='members_list_overlay'),
                                                                           dynamic=False, cls='board-labels-overlay'))

        self.comp_members = component.Component(self)

        # Icons for the toolbar
        self.icons = {'add_list': component.Component(Icon('icon-plus', _('Add list'))),
                      'edit_desc': component.Component(Icon('icon-pencil', _('Edit board description'))),
                      'preferences': component.Component(Icon('icon-cog', _('Preferences'))),
                      'export': component.Component(Icon('icon-download3', _('Export board'))),
                      'save_template': component.Component(Icon('icon-insert-template', _('Save as template'))),
                      'archive': component.Component(Icon('icon-bin', _('Archive board'))),
                      'leave': component.Component(Icon('icon-exit', _('Leave this board'))),
                      'history': component.Component(Icon('icon-history', _("Action log"))),
                      }

        # Title component
        self.title = component.Component(
            title.EditableTitle(self.get_title)).on_answer(self.set_title)

    @classmethod
    def get_id_by_uri(cls, uri):
        board = DataBoard.get_by_uri(uri)
        board_id = None
        if board is not None:
            board_id = board.id
        return board_id

    @classmethod
    def exists(cls, **kw):
        return DataBoard.exists(**kw)

    def __eq__(self, other):
        return isinstance(other, Board) and self.id == other.id

    # Main menu actions
    def add_list(self):
        new_column_editor = column.NewColumnEditor(len(self.columns) - 1)
        answer = self.modal.call(popin.Modal(new_column_editor, force_refresh=True))
        if answer:
            index, title, nb_cards = answer
            self.create_column(index, title, nb_cards if nb_cards else None)

    def edit_description(self):
        description_editor = BoardDescription(self.get_description())
        answer = self.modal.call(popin.Modal(description_editor))
        if answer is not None:
            self.set_description(answer)

    def save_template(self, comp):
        save_template_editor = SaveTemplateTask(self.get_title(),
                                                self.get_description(),
                                                partial(self.save_as_template, comp))
        self.modal.call(popin.Modal(save_template_editor))

    def show_actionlog(self):
        self.modal.call(popin.Modal(self.action_log))

    def show_preferences(self):
        preferences = BoardConfig(self)
        self.modal.call(popin.Modal(preferences, force_refresh=True))

    def save_as_template(self, comp, title, description, shared):
        data = (title, description, shared)
        return self.emit_event(comp, events.NewTemplateRequested, data)

    def copy(self, owner):
        """
        Create a new board that is a copy of self, without the archive.
        Children must be loaded.
        """
        new_data = self.data.copy()
        if self.data.background_image:
            new_data.background_image = self.assets_manager.copy(self.data.background_image)
        new_board = self._services(Board, new_data.id, self.app_title, self.app_banner, self.theme,
            self.card_extensions, load_children=False, data=new_data)
        new_board.add_member(owner, 'manager')

        assert(self.columns or self.data.is_template)
        cols = [col() for col in self.columns if not col().is_archive]
        for column in cols:
            new_column = new_board.create_column(-1, column.get_title())
            new_column.update(column)

        return new_board

    def on_event(self, comp, event):
        result = None
        if event.is_(events.ColumnDeleted):
            # actually delete the column
            result = self.delete_column(event.data)
        elif event.is_(events.CardArchived):
            result = self.archive_cards([event.emitter], event.last_relay)
        elif event.is_(events.SearchIndexUpdated):
            result = self.card_filter.reload_search()

        return result

    def switch_view(self):
        self.model = 'calendar' if self.model == 'columns' else 'columns'

    def load_children(self):
        columns = []
        for c in self.data.columns:
            col = self._services(
                column.Column, c.id, self, self.card_extensions,
                self.action_log, self.card_filter, data=c)
            if col.is_archive:
                self.archive_column = col
            columns.append(component.Component(col))

        self.columns = columns

    def increase_version(self):
        self.version += 1
        self.data.increase_version()
        return self.refresh_on_version_mismatch()

    def refresh_on_version_mismatch(self):
        refresh = False
        if self.data.version - self.version != 0:
            self.refresh()  # when does that happen?
            self.version = self.data.version
            refresh = True
        return refresh

    def refresh(self):
        log.info('sync')
        self._data = None
        self.load_children()

    @property
    def all_members(self):
        return self.managers + self.members + self.pending

    def update_members(self):
        """Update members section

        Recalculate members + managers + pending
        Recreate overlays
        """
        data = self.data
        #FIXME: use Membership components
        managers = []
        simple_members = []
        for manager, memberships in itertools.groupby(data.board_members,
                                                      lambda item: item.manager):
            # we use extend because the board_members are not reordered in case of change
            if manager:
                managers.extend([membership.user for membership in memberships])
            else:
                simple_members.extend([membership.user for membership in memberships])
        simple_members.sort(key=lambda m: (m.fullname, m.email))
        self.members = [component.Component(BoardMember(usermanager.UserManager.get_app_user(data=member), self, 'member'))
                        for member in simple_members]
        self.managers = [component.Component(BoardMember(usermanager.UserManager.get_app_user(data=member), self, 'manager' if len(managers) != 1 else 'last_manager'))
                         for member in managers]
        self.pending = [component.Component(BoardMember(PendingUser(token.token), self, 'pending'))
                        for token in data.pending]

    def set_title(self, title):
        """Set title

        In:
            - ``title`` -- new title
        """
        self.data.title = title

    def get_title(self):
        """Get title

        Return :
            - the board title
        """
        return self.data.title

    def mark_as_template(self, template=True):
        self.data.is_template = template

    def count_columns(self):
        """Return the number of columns
        (used in unit tests only)
        """
        return len(self.columns)

    @security.permissions('edit')
    def create_column(self, index, title, nb_cards=None):
        """Create a new column in the board

        In:
            - ``index`` -- the position of the column as an integer
            - ``title`` -- the title of the new column
            - ``nb_cards`` -- the number of maximun cards on the colum
        """
        if index < 0:
            index = index + len(self.columns) + 1
        if title == '':
            return False
        col = self.data.create_column(index, title, nb_cards)
        col_obj = self._services(
            column.Column, col.id, self,
            self.card_extensions, self.action_log, self.card_filter)
        self.columns.insert(
            index, component.Component(col_obj))
        self.increase_version()
        return col_obj

    @security.permissions('edit')
    def delete_column(self, col_comp):
        """Delete a board's column

        In:
            - ``id_`` -- the id of the column to delete
        """
        self.columns.remove(col_comp)
        self.data.delete_column(col_comp().data)
        col_comp().delete()
        self.increase_version()
        return popin.Empty()

    @security.permissions('edit')
    def update_card_position(self, data):
        data = json.loads(data)

        cols = {}
        for col in self.columns:
            cols[col().id] = (col(), col)

        orig, __ = cols[data['orig']]

        dest, dest_comp = cols[data['dest']]
        card_comp = None
        try:
            card_comp = orig.remove_card_by_id(data['card'])
            accepted = dest.insert_card_comp(dest_comp, data['index'], card_comp)
        except AttributeError:
            # one of the columns does not exist anymore
            # stop processing, let the refresh do the rest
            log.warning('attempt to move card between at least one missing column')
            if card_comp:
                orig.append_card(card_comp())
            return
        if accepted:
            card = card_comp()
            values = {'from': orig.get_title(),
                      'to': dest.get_title(),
                      'card': card.get_title()}
            self.action_log.for_card(card).add_history(
                security.get_user(),
                u'card_move', values)
            # reindex it in case it has been moved to the archive column
            card.add_to_index(self.search_engine, self.id, update=True)
            self.search_engine.commit()
            session.flush()
        else:
            orig.append_card(card_comp())

    @security.permissions('edit')
    def update_column_position(self, data):
        data = json.loads(data)
        cols = []
        found = None
        for col in self.columns:
            if col().id == data['list']:
                found = col
            else:
                cols.append(col)
        cols.insert(data['index'], found)
        for i, col in enumerate(cols):
            col().change_index(i)
        self.columns = cols
        session.flush()

    @property
    def visibility(self):
        return self.data.visibility

    @property
    def is_open(self):
        return (self.visibility == BOARD_PUBLIC or self.visibility == BOARD_SHARED)

    def set_visibility(self, visibility):
        """Changes board visibility

        If new visibility is "Member" and comments/votes permissions
        are in "Public" changes them to "Members"

        In:
         - ``visibility`` -- an integer, new visibility (Private or Public)
        """
        if self.comments_allowed == COMMENTS_PUBLIC:
            # If comments are PUBLIC that means the board was PUBLIC and
            # go to PRIVATE. That's why we don't test the visibility
            # input variable
            self.allow_comments(COMMENTS_MEMBERS)
        if self.votes_allowed == VOTES_PUBLIC:
            self.allow_votes(VOTES_MEMBERS)
        self.data.visibility = visibility

    @property
    def archived(self):
        return self.data.archived

    @property
    def show_archive(self):
        return self.data.show_archive

    @show_archive.setter
    def show_archive(self, value):
        self.data.show_archive = value
        self.card_filter.exclude_archived(not value)

    def archive_cards(self, cards, from_column):
        """Archive card

        In:
            - ``cards`` -- cards to archive, from the same column
        """
        for card in cards:
            self.archive_column.append_card(card)
            values = {'column_id': from_column.id, 'column': from_column.get_title(),
                      'card': card.get_title()}
            card.action_log.add_history(security.get_user(), u'card_archive', values)
            # reindex it
            card.add_to_index(self.search_engine, self.id, update=True)
        self.search_engine.commit(True)
        self.card_filter.reload_search()
        self.increase_version()

    ####### For future board extension

    @property
    def weighting_cards(self):
        return self.data.weighting_cards

    def activate_weighting(self, weighting_type):
        self.data.weighting_cards = weighting_type
        if weighting_type != WEIGHTING_FREE:
            # reinitialize card weights?
            self.data.reset_card_weights()

    @property
    def weights(self):
        if not self.data.weights:
            self.data.weights = '0, 1, 2, 3, 5, 8, 13'
        return self.data.weights

    @weights.setter
    def weights(self, weights):
        self.data.weights = weights

    def total_weight(self):
        return self.data.total_weight()

    ######################

    def delete_clicked(self, comp):
        return self.emit_event(comp, events.BoardDeleted)

    def delete(self):
        """Deletes the board.
           Children must be loaded.
        """
        assert(self.columns)  # at least, contains the archive
        for column in self.columns:
            column().delete(purge=True)
        self.data.delete_history()
        self.data.delete_members()
        if self.data.background_image:
            self.assets_manager.delete(self.data.background_image)
        session.refresh(self.data)
        self.data.delete()

        return True

    def archive(self, comp=None):
        """Archive the board
        """
        self.data.archived = True
        if comp:
            self.emit_event(comp, events.BoardArchived)
        return True

    def restore(self, comp=None):
        """Unarchive the board
        """
        self.data.archived = False
        if comp:
            self.emit_event(comp, events.BoardRestored)
        return True

    def export(self):
        return ExcelExport(self).download()

    @property
    def labels(self):
        """Returns the labels associated with the board
        """
        return [self._services(Label, data) for data in self.data.labels]

    @property
    def data(self):
        """Return the board object from the database
        PRIVATE
        """
        if self._data is None:
            self._data = DataBoard.get(self.id)
        return self._data

    def __getstate__(self):
        self._data = None
        return self.__dict__

    def allow_comments(self, v):
        """Changes permission to add comments

        In:
            - ``v`` -- a integer (see security.py for authorized values)
        """
        self.data.comments_allowed = v

    def allow_votes(self, v):
        """Changes permission to vote

        In:
            - ``v`` -- a integer (see security.py for authorized values)
        """
        self.data.votes_allowed = v

    @property
    def comments_allowed(self):
        return self.data.comments_allowed

    @property
    def votes_allowed(self):
        return self.data.votes_allowed

    # Callbacks for BoardDescription component
    def get_description(self):
        return self.data.description

    def set_description(self, value):
        self.data.description = value


    ##################
    # Member methods
    ##################

    def leave(self, comp=None):
        """Children must be loaded."""
        # FIXME: all member management function should live in another component than Board.
        user = security.get_user()
        for member in self.members:
            m_user = member().user().data
            if (m_user.username, m_user.source) == (user.data.username, user.data.source):
                board_member = member()
                break
        else:
            board_member = None
        self.data.remove_member(board_member.data)
        if comp:
            self.emit_event(comp, events.BoardLeft)
        return True

    def last_manager(self, member):
        """Return True if member is the last manager of the board

        In:
         - ``member`` -- member to test
        Return:
         - True if member is the last manager of the board
        """
        return member.role == 'manager' and len(self.managers) == 1

    def has_member(self, user):
        """Return True if user is member of the board

        In:
         - ``user`` -- user to test (User instance)
        Return:
         - True if user is member of the board
        """
        return self.data.has_member(user.data)

    def has_manager(self, user):
        """Return True if user is manager of the board

        In:
         - ``user`` -- user to test (User instance)
        Return:
         - True if user is manager of the board
        """
        return self.data.has_manager(user.data)

    def add_member(self, new_member, role='member'):
        """ Add new member to the board

        In:
         - ``new_member`` -- user to add
         - ``role`` -- role's member (manager or member)
        """
        self.data.add_member(new_member.data, role)

    def remove_pending(self, member):
        # remove from pending list
        self.pending = [p for p in self.pending if p() != member]

        # remove invitation
        self.remove_invitation(member.username)

    def remove_manager(self, manager):
        # remove from managers list
        self.managers = [p for p in self.managers if p() != manager]
        # remove manager from data part
        self.data.remove_member(manager.data)

    def remove_member(self, member):
        # remove from members list
        self.members = [p for p in self.members if p() != member]
        # remove member from data part
        self.data.remove_member(member.data)

    def remove_board_member(self, member):
        """Remove member from board

        Remove member from board. If member is PendingUser then remove
        invitation.

        Children must be loaded for propagation to the cards.

        In:
            - ``member`` -- Board Member instance to remove
        """
        if self.last_manager(member):
            # Can't remove last manager
            raise exceptions.KanshaException(_("Can't remove last manager"))

        log.info('Removing member %s' % (member,))
        remove_method = {'pending': self.remove_pending,
                         'manager': self.remove_manager,
                         'member': self.remove_member}
        remove_method[member.role](member)

    def change_role(self, member, new_role):
        """Change member's role

        In:
            - ``member`` -- Board member instance
            - ``new_role`` -- new role
        """
        log.info('Changing role of %s to %s' % (member, new_role))
        if self.last_manager(member):
            raise exceptions.KanshaException(_("Can't remove last manager"))

        self.data.change_role(member.data, new_role)
        self.update_members()

    def remove_invitation(self, email):
        """ Remove invitation

        In:
         - ``email`` -- guest email to invalidate
        """
        for token in self.data.pending:
            if token.username == email:
                token.delete()
                session.flush()
                break

    def invite_members(self, emails, application_url):
        """Invite somebody to this board,

        Create token used in invitation email.
        Store email in pending list.

        Params:
            - ``emails`` -- list of emails
        """
        for email in set(emails):
            # If user already exists add it to the board directly or invite it otherwise
            invitation = forms.EmailInvitation(self.app_title, self.app_banner, self.theme, email, security.get_user().data, self.data, application_url)
            invitation.send_email(self.mail_sender)

    def resend_invitation(self, pending_member, application_url):
        """Resend an invitation,

        Resend invitation to the pending member

        In:
            - ``pending_member`` -- Send invitation to this user (PendingMember instance)
        """
        email = pending_member.username
        invitation = forms.EmailInvitation(self.app_title, self.app_banner, self.theme, email, security.get_user().data, self.data, application_url)
        invitation.send_email(self.mail_sender)
        # re-calculate pending
        self.pending = [component.Component(BoardMember(PendingUser(token.token), self, "pending"))
                        for token in set(self.data.pending)]

################

    def autocomplete_method(self, v):
        """ Method called by autocomplete component.

        This method is called when you search a user on the add member
        overlay int the field autocomplete

        In:
            - ``v`` -- first letters of the username
        Return:
            - list of user (User instance)
        """
        users = usermanager.UserManager.search(v)
        results = []
        for user in users:
            if user.is_validated() and user.email not in [m().email for m in self.all_members]:
                results.append(user)
        return results

    def get_last_activity(self):
        return self.action_log.get_last_activity()

    def get_available_user_ids(self):
        """Return list of member

        Return:
            - list of members
        """
        return set(dbm.user.id for dbm in self.data.board_members)

    def set_background_image(self, new_file):
        """Set the board's background image
        In:
            - ``new_file`` -- the background image (FieldStorage)
        Return:
            nothing
        """
        if new_file is not None:
            fileid = self.assets_manager.save(new_file.file.read(),
                                              metadata={'filename': new_file.filename,
                                                        'content-type': new_file.type})
            self.data.background_image = fileid
        else:
            self.data.background_image = None

    def set_background_position(self, position):
        self.data.background_position = position

    @property
    def background_image_url(self):
        img = self.data.background_image
        try:
            return self.assets_manager.get_image_url(img, include_filename=False) if img else None
        except IOError:
            log.warning('Missing background %r for board %r', img, self.id)
            return None

    @property
    def background_image_position(self):
        return self.data.background_position or 'center'

    @property
    def title_color(self):
        return self.data.title_color

    def set_title_color(self, value):
        self.data.title_color = value or u''

    @classmethod
    def get_all_boards(cls, user, app_title, app_banner, theme, card_extensions,
                       services_service, load_children=False):
        """Return all boards the user is member of."""
        return [services_service(cls, data.id, app_title, app_banner, theme, card_extensions,
                                 data=data, load_children=load_children)
                for data in DataBoard.get_all_boards(user.data)]

    @classmethod
    def get_shared_boards(cls, app_title, app_banner, theme, card_extensions,
                          services_service, load_children=False):
        """Return all boards the user is member of."""
        return [services_service(cls, data.id, app_title, app_banner, theme, card_extensions,
                                 data=data, load_children=load_children)
                for data in DataBoard.get_shared_boards()]

    @staticmethod
    def get_templates_for(user):
        return DataBoard.get_templates_for(user.data, BOARD_PUBLIC)


# TODO: move this to board extension
@when(common.Rules.has_permission, "user and perm == 'Add Users' and isinstance(subject, Board)")
def has_permission_Board_add_users(self, user, perm, board):
    """Test if users is one of the board's managers, if he is he can add new user to the board"""
    return board.has_manager(user)

################


class Icon(object):

    def __init__(self, icon, title=None):
        """Create icon object

        In:
          - ``icon`` -- icon class name (use icomoon custom font)
          - ``title`` -- icon title (and alt)
        """
        self.icon = icon
        self.title = title

################


class BoardDescription(object):

    """Description component
    """

    def __init__(self, description):
        """Initialization

        In:
            - ``description`` -- callable that returns the description.
        """
        self.description = var.Var(description)

    def commit(self, comp):
        description = self.description().strip()
        if description:
            description = validator.clean_text(description)
        comp.answer(description)

    def cancel(self, comp):
        comp.answer(None)


class BoardMember(object):

    def __init__(self, user, board, role):
        self.user = component.Component(user)
        self.role = role
        self.board = board

    @property
    def username(self):
        return self.user().username

    @property
    def fullname(self):
        return self.user().fullname

    @property
    def email(self):
        return self.user().email

    @property
    def data(self):
        return self.user().data

    def dispatch(self, action, application_url):
        if action == 'remove':
            self.board.remove_board_member(self)
        elif action == 'toggle_role':
            self.board.change_role(self, 'manager' if self.role == 'member' else 'member')
        elif action == 'resend':
            self.board.resend_invitation(self, application_url)



# -*- coding:utf-8 -*-
# --
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.i18n import _
from nagare import ajax, component, presentation, security, var

from kansha.toolbox import popin

from .comp import CardsCounter, Column, NewColumnEditor


@presentation.render_for(Column)
def render(self, h, comp, *args):
    """Render the column"""

    column_class = 'span-auto list'
    if self.is_archive:
        column_class += ' archive'
    with h.div(class_=column_class, id=self.id, ):
        h << comp.render(h.AsyncRenderer(), 'content')
    return h.root


@presentation.render_for(Column, 'content')
def render_content(self, h, comp, model):
    h << comp.render(h.AsyncRenderer(), 'header')
    h << comp.render(h, 'body')
    h << self.card_counter.render(h, 'footer')
    h << component.Component(self.card_filter, 'footer')
    return h.root


@presentation.render_for(Column, 'calendar')
def render_column_calendar(self, h, comp, *args):
    return [card.render(h.AsyncRenderer(), 'calendar') for card in self.cards]


@presentation.render_for(Column, 'new')
def render_column_new(self, h, comp, *args):
    return h.div(comp.becomes(self, 'dnd'), class_='new')


@presentation.render_for(Column, model='dropdown')
def render_column_dropdown(self, h, comp, *args):
    """Render the column menu"""
    with h.div(class_="dropdown menu"):
        with h.ul:
            if not self.is_archive:
                with h.li:
                    onclick = (
                        u"if (confirm(%(message)s)){"
                        u" window.location='%(callback)s';"
                        u"}" %
                        {
                            'message': ajax.py2js(
                                _(u'The list will be deleted. Are you sure?')
                            ).decode('UTF-8'),
                            'callback': h.SyncRenderer().a.action(
                                self.actions, 'delete', comp
                            ).get('href')
                        }
                    )
                    h << h.a(_(u'Delete this list'), onclick=onclick)
                if self.cards:
                    with h.li:
                        onclick = (
                            u"if (confirm(%(message)s)){"
                            u" window.location='%(callback)s';"
                            u"}" %
                            {
                                'message': ajax.py2js(
                                    _(u'All the cards will be archived. Are you sure?')
                                ).decode('UTF-8'),
                                'callback': h.SyncRenderer().a.action(
                                    self.actions, 'empty', comp
                                ).get('href')
                            }
                        )
                        h << h.a(_(u'Empty this list'), onclick=onclick)
                h << self.card_counter.render(h, 'menu-entry')
            elif self.cards:
                with h.li:
                    onclick = "if (confirm(%(message)s)){window.location='%(purge_func)s';}" % {
                        'message': ajax.py2js(
                            _(u'All cards will be deleted. Are you sure?')
                        ).decode('UTF-8'),
                        'purge_func': h.SyncRenderer().a.action(
                            self.actions, 'purge', comp
                        ).get('href')
                    }
                    h << h.a(_('Purge the cards'), onclick=onclick)

    return h.root


@presentation.render_for(Column, model='header')
def render_column_header(self, h, comp, *args):
    with h.div(class_='list-header', id=self.id + '_header'):
        h << h.a(class_='hidden', id=self.id + '_refresh').action(ajax.Update())
        with h.div(class_='list-title'):
            with h.div(class_='title'):
                h << self.title.render(h.AsyncRenderer(), 0 if security.has_permissions('edit', self) and not self.is_archive else 'readonly')
        h << self.card_counter.render(h, 'header')
        with h.div(class_='list-actions with-dropdown'):
            if security.has_permissions('edit', self):
                h << h.a(h.i(class_='icon-dot-menu'),
                         href='#', class_="toggle-dropdown",
                         onclick="YAHOO.kansha.app.toggleMenu(this)") << ' '
                h << comp.render(h, 'dropdown')
    return h.root


@presentation.render_for(Column, model='title')
def render_column_title(self, h, comp, *args):
    with h.div(class_='title'):
        h << self.title.render(h, 'readonly')
    return h.root


@presentation.render_for(Column, 'dnd')
def render_column_dnd(self, h, comp, *args):
    """DnD wrapper for column"""
    h << comp.render(h, None)
    h << h.script(
        "YAHOO.util.Event.onDOMReady(function() {"
        "  YAHOO.kansha.dnd.initList(%(list_id)s);"
        "})" % {'list_id': ajax.py2js(self.id)}
    )
    return h.root


@presentation.render_for(Column, 'body')
def render_column_body(self, h, comp, *args):
    model = 'dnd' if security.has_permissions('edit', self) else "no_dnd"
    id_ = h.generate_id()
    with h.div(class_='list-body', id=id_):
        h << [card.on_answer(self.handle_event, comp).render(h, model=model) for card in self.cards]
        h << h.script("YAHOO.kansha.dnd.initTargetCard(%s)" % ajax.py2js(id_))
        # h << self.card_counter.render(h, 'body')
    kw = {}
    if not security.has_permissions('edit', self):
        kw['style'] = 'width: 0px'
    with h.div(class_='list-footer', id=self.id + '_footer', **kw):
        # This HACK is to force nagare to insert its div, otherwise it breaks all the layout.
        if self.is_archive:
            h << {'style': 'display: none'}
        if security.has_permissions('edit', self):
            h << h.div(self.new_card.on_answer(self.ui_create_card, comp))

    return h.root


@presentation.render_for(NewColumnEditor)
def render_NewColumnEditor(self, h, comp, *args):
    """Render column creator"""
    h << h.h2(_(u'Add list'))
    with h.form:
        with h.div:
            id_ = h.generate_id()
            h << h.label(_('Name'), for_=id)
            h << h.input(id=id_, type='text', value=self.title(),
                         placeholder=_('List title'),
                         autofocus=True).error(self.title.error).action(self.title)

        with h.div:
            id_ = h.generate_id()
            h << h.label(_('Position'), for_=id_)
            with h.select(id=id_).error(self.index.error).action(self.index):
                for i in xrange(1, self.columns_count + 2):
                    h << h.option(i, value=i - 1).selected(i)

        # FIXME
        with h.div:
            id_ = h.generate_id()
            h << h.label(_('Number max of cards'), id_=id_)
            h << h.input(id=id_, type='text', value=self.nb_cards()).error(self.nb_cards.error).action(self.nb_cards)
            h << h.script(
                """YAHOO.util.Event.on(%s, 'keyup', function (e) {
                        var result =this.value.replace(/[^0-9]/g, '')
                        if (this.value !=result) {
                               this.value = result;
                            }
                 })""" % ajax.py2js(id_)
            )

        with h.div(class_='buttons'):
            h << h.button(_('Add'), class_=('btn btn-primary')).action(self.commit, comp)
            h << ' '
            h << h.a(_('Cancel'), class_='btn').action(self.cancel, comp)
    return h.root


@presentation.render_for(CardsCounter)
def render_CardsCounter(self, h, comp, *args):
    with h.div(class_='list-counter'):
        self.error = None
        visibility = ' hidden'
        if self.column.nb_max_cards:
            visibility = '' if self.check_add() else ' limitReached'
        with h.div(class_='cardCounter' + visibility, id=self.id):
            with h.a().action(comp.call, self, 'edit'):
                h << self.column.count_cards << '/' << (self.column.nb_max_cards or 0)
    h << h.script(
        "YAHOO.kansha.app.saveLimit(%(list_id)s, %(limit)s);"
        "YAHOO.kansha.app.countCards(%(list_id)s);" %
        {
            'list_id': ajax.py2js(self.column.id),
            'limit': ajax.py2js(self.column.nb_max_cards or 0)
        }
    )
    return h.root


@presentation.render_for(CardsCounter, 'header')
def render_CardsCounter_header(self, h, comp, model):
    h << self.editable_counter
    return h.root


@presentation.render_for(CardsCounter, 'menu-entry')
def render_CardsCounter_menu(self, h, comp, model):
    with h.li:
        h << h.a(_(u'Set card limit')).action(self.editable_counter.call, self, 'edit')
    return h.root


@presentation.render_for(CardsCounter, 'body')
def render_CardsCounter_body(self, h, comp, model):
    with h.div(class_='no-drop'):
        h << h.i(class_='icon-blocked huge') << h.br
        h << _(u"This list already holds its maximum amount of cards")
    h << h.script("YAHOO.kansha.app.countCards(%s)" % ajax.py2js(self.column.id))
    return h.root


@presentation.render_for(CardsCounter, 'footer')
def render_CardsCounter_footer(self, h, comp, model):
    h << h.script(
        "YAHOO.kansha.app.countCards(%(list_id)s);" %
        {
            'list_id': ajax.py2js(self.column.id),
        }
    )
    return h.root


@presentation.render_for(CardsCounter, model='edit')
def render_CardsCounter_edit(self, h, comp, *args):
    """Render the title of the associated object"""
    text = var.Var(self.text)
    with h.div(class_='list-counter'):
        with h.div(class_='cardCounter'):
            with h.form(onsubmit='return false;'):
                action = h.input(type='submit').action(lambda: self.validate(text(), comp)).get('onclick')
                id_ = h.generate_id()
                h << h.input(id=id_, type='text', value=self.column.nb_max_cards or '', onblur=action).action(text)
                h << h.script(
                    """YAHOO.util.Event.on(%s, 'keyup', function (e) {
                            if (e.keyCode == 13) {
                                e.preventDefault();
                                this.blur();
                            }
                            var result = this.value.replace(/[^0-9]/g, '')
                            if (this.value !=result) {
                               this.value = result;
                            }
                     });""" % ajax.py2js(id_)
                )
                h << h.script(
                    "YAHOO.kansha.app.selectElement(%s);" % ajax.py2js(id_)
                )
        if self.error is not None:
            with h.div(class_='nagare-error-message'):
                h << self.error
    return h.root



# -*- coding:utf-8 -*-
#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--
import random
from datetime import datetime, timedelta

from nagare.namespaces import xhtml
from nagare import component, i18n

from .comp import User
from .models import DataUser
from kansha.toolbox import autocomplete


class UserManager(object):

    @classmethod
    def get_app_user(cls, username=None, data=None):
        """Return User instance"""
        if not data:
            data = cls.get_by_username(username)
        else:
            username = data.username
        if data.source != 'application':
            # we need to set a passwd for nagare auth
            user = User(username, 'passwd', data=data)
        else:
            user = User(username, data=data)
        return user

    @staticmethod
    def search(value):
        """Return all users which name begins by value"""
        return DataUser.search(value)

    @staticmethod
    def get_all_users(hours=0):
        """Return all data users if `hours` is 0 or just those who have registrated
        for the last hours."""
        q = DataUser.query
        if hours:
            since = datetime.utcnow() - timedelta(hours=hours)
            q = q.filter(DataUser.registration_date >= since)
        return q.all()

    @staticmethod
    def get_by_username(username):
        return DataUser.get_by_username(username)

    @staticmethod
    def get_by_email(email):
        return DataUser.get_by_email(email)

    @staticmethod
    def get_unconfirmed_users(before_date):
        """Return unconfirmed user

        Return all user which have'nt email validated before a date
        In:
            - ``before_date`` -- limit date
        Return:
            - list of DataUser instance
        """
        return DataUser.get_unconfirmed_users(before_date)

    @staticmethod
    def get_confirmed_users():
        """Return confirmed user

        Return all user which have email validated

        Return:
            - list of DataUser instance
        """
        return DataUser.get_confirmed_users()

    def create_user(self, username, password, fullname, email,
                    source=u'application', picture=None):
        from kansha.authentication.database import forms
        user = DataUser(username, password, fullname, email,
                        source, picture, language=i18n.get_locale().language)
        token_gen = forms.TokenGenerator(email, u'invite board')
        for token in token_gen.get_tokens():
            if token_gen.check_token(token.token) and token.board:
                token.board.add_member(user)
            token_gen.reset_token(token.token)
        return user


###### TODO: Move the defintions below somewhere else ##########

class NewMember(object):

    """New Member Class"""

    def __init__(self, autocomplete_method):
        """Init method

        In:
            - ``autocomplete_methode`` -- method called for autocomplete
        """
        self.text_id = 'new_members_' + str(random.randint(10000000, 99999999))
        self.autocomplete = component.Component(
            autocomplete.Autocomplete(self.text_id, self.autocompletion, delimiter=','))
        self.autocomplete_method = autocomplete_method

    def autocompletion(self, value, static_url):
        """Return users with email which match with value.

        Method called by autocomplete. This method returns a list of tuples.
        First element of tuple is the email's user. The second is a HTML
        fragment to display in the list of matches.

        In:
         - ``value`` -- first letters of user email
        Return:
         - list of tuple (email, HTML string)
        """
        # FIXME: dont access data attributes!
        h = xhtml.Renderer(static_url=static_url)
        return [(u.email or u.email_to_confirm, component.Component(UserManager.get_app_user(data=u)).render(h, "search").write_htmlstring())
                for u in self.autocomplete_method(value)]


# class AddMembers(NewMember):

#     pass



"""new_background_positions

Revision ID: 5791b368ac26
Revises: 443501b69d16
Create Date: 2017-04-27 11:03:20.788762

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = '5791b368ac26'
down_revision = '443501b69d16'


def upgrade():
    op.alter_column('board', 'background_position', nullable=True, server_default=None)
    board = sa.Table(
        'board',
        sa.MetaData(),
        sa.Column('background_position', sa.Unicode),
    )
    bind = op.get_bind()
    bind.execute(
        board.update().where(
            board.c.background_position == u'repeat'
        ).values(background_position=u'tile')
    )
    bind.execute(
        board.update().where(
            board.c.background_position == u'cover'
        ).values(background_position=u'fill')
    )


def downgrade():
    pass



#--
# Copyright (c) 2012-2014 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
#--

from nagare.services import plugin

from kansha.events import EventHandlerMixIn


class CardExtension(plugin.Plugin, EventHandlerMixIn):
    CATEGORY = 'card-extension'

    def __init__(self, card, action_log, configurator=None):
        self.card = card
        self.action_log = action_log
        self.configurator = configurator

    @staticmethod
    def get_excel_title():
        '''If the extension is exportable in Excel, return its column name'''
        return None

    @staticmethod
    def get_schema_def():
        '''If the extension has to be indexed for it to be used in search engine, return some schema field
        ie: return schema.Text for a text field'''
        return None

    def update_document(self, document):
        '''Add extension value to document that will be indexed'''
        pass

    def delete(self):
        '''Happens when a card is deleted, use it to clean up files for example'''
        pass

    def update(self, other):
        '''Copy state and data from other on self.'''
        pass

    def new_card_position(self, value):
        '''Happens when a card is moved on the board'''
        pass

    def has_permission_on_card(self, user, perm):
        return True
